"""
Tests for the win_pkg module
"""

import salt.modules.config as config
import salt.modules.pkg_resource as pkg_resource
import salt.modules.win_pkg as win_pkg
import salt.utils.data
import salt.utils.platform
import salt.utils.win_reg as win_reg
from salt.exceptions import MinionError
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase, skipIf


@skipIf(not salt.utils.platform.is_windows(), "Must be on Windows!")
class WinPkgInstallTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.win_pkg
    """

    def setup_loader_modules(self):
        pkg_info = {
            "3.03": {
                "full_name": "Nullsoft Install System",
                "installer": "http://download.sourceforge.net/project/nsis/NSIS%203/3.03/nsis-3.03-setup.exe",
                "install_flags": "/S",
                "uninstaller": "%PROGRAMFILES(x86)%\\NSIS\\uninst-nsis.exe",
                "uninstall_flags": "/S",
                "msiexec": False,
                "reboot": False,
            },
            "3.02": {
                "full_name": "Nullsoft Install System",
                "installer": "http://download.sourceforge.net/project/nsis/NSIS%203/3.02/nsis-3.02-setup.exe",
                "install_flags": "/S",
                "uninstaller": "%PROGRAMFILES(x86)%\\NSIS\\uninst-nsis.exe",
                "uninstall_flags": "/S",
                "msiexec": False,
                "reboot": False,
            },
        }

        return {
            win_pkg: {
                "_get_latest_package_version": MagicMock(return_value="3.03"),
                "_get_package_info": MagicMock(return_value=pkg_info),
                "__salt__": {
                    "pkg_resource.add_pkg": pkg_resource.add_pkg,
                    "pkg_resource.parse_targets": pkg_resource.parse_targets,
                    "pkg_resource.sort_pkglist": pkg_resource.sort_pkglist,
                    "pkg_resource.stringify": pkg_resource.stringify,
                    "config.valid_fileproto": config.valid_fileproto,
                },
                "__utils__": {
                    "reg.key_exists": win_reg.key_exists,
                    "reg.list_keys": win_reg.list_keys,
                    "reg.read_value": win_reg.read_value,
                    "reg.value_exists": win_reg.value_exists,
                },
            },
            pkg_resource: {"__grains__": {"os": "Windows"}},
        }

    def test_pkg__get_reg_software(self):
        result = win_pkg._get_reg_software()
        self.assertTrue(isinstance(result, dict))
        found_python = False
        search = "Python 3"
        for key in result:
            if search in key:
                found_python = True
        self.assertTrue(found_python)

    def test_pkg_install_not_found(self):
        """
        Test pkg.install when the Version is NOT FOUND in the Software
        Definition
        """
        ret_reg = {"Nullsoft Install System": "3.03"}
        # The 2nd time it's run with stringify
        se_list_pkgs = {"nsis": ["3.03"]}
        with patch.object(
            win_pkg, "list_pkgs", return_value=se_list_pkgs
        ), patch.object(win_pkg, "_get_reg_software", return_value=ret_reg):
            expected = {"nsis": {"not found": "3.01"}}
            result = win_pkg.install(name="nsis", version="3.01")
            self.assertDictEqual(expected, result)

    def test_pkg_install_rollback(self):
        """
        test pkg.install rolling back to a previous version
        """
        ret_reg = {"Nullsoft Install System": "3.03"}
        # The 2nd time it's run, pkg.list_pkgs uses with stringify
        se_list_pkgs = [{"nsis": ["3.03"]}, {"nsis": "3.02"}]
        with patch.object(win_pkg, "list_pkgs", side_effect=se_list_pkgs), patch.object(
            win_pkg, "_get_reg_software", return_value=ret_reg
        ), patch.dict(
            win_pkg.__salt__, {"cp.is_cached": MagicMock(return_value=False)}
        ), patch.dict(
            win_pkg.__salt__,
            {"cp.cache_file": MagicMock(return_value="C:\\fake\\path.exe")},
        ), patch.dict(
            win_pkg.__salt__, {"cmd.run_all": MagicMock(return_value={"retcode": 0})}
        ):
            expected = {"nsis": {"new": "3.02", "old": "3.03"}}
            result = win_pkg.install(name="nsis", version="3.02")
            self.assertDictEqual(expected, result)

    def test_pkg_install_existing(self):
        """
        test pkg.install when the package is already installed
        no version passed
        """
        ret_reg = {"Nullsoft Install System": "3.03"}
        # The 2nd time it's run, pkg.list_pkgs uses with stringify
        se_list_pkgs = {"nsis": ["3.03"]}
        with patch.object(
            win_pkg, "list_pkgs", return_value=se_list_pkgs
        ), patch.object(win_pkg, "_get_reg_software", return_value=ret_reg), patch.dict(
            win_pkg.__salt__, {"cp.is_cached": MagicMock(return_value=False)}
        ), patch.dict(
            win_pkg.__salt__,
            {"cp.cache_file": MagicMock(return_value="C:\\fake\\path.exe")},
        ), patch.dict(
            win_pkg.__salt__, {"cmd.run_all": MagicMock(return_value={"retcode": 0})}
        ):
            expected = {}
            result = win_pkg.install(name="nsis")
            self.assertDictEqual(expected, result)

    def test_pkg_install_existing_with_version(self):
        """
        test pkg.install when the package is already installed
        A version is passed
        """
        ret_reg = {"Nullsoft Install System": "3.03"}
        # The 2nd time it's run, pkg.list_pkgs uses with stringify
        se_list_pkgs = {"nsis": ["3.03"]}
        with patch.object(
            win_pkg, "list_pkgs", return_value=se_list_pkgs
        ), patch.object(win_pkg, "_get_reg_software", return_value=ret_reg), patch.dict(
            win_pkg.__salt__, {"cp.is_cached": MagicMock(return_value=False)}
        ), patch.dict(
            win_pkg.__salt__,
            {"cp.cache_file": MagicMock(return_value="C:\\fake\\path.exe")},
        ), patch.dict(
            win_pkg.__salt__, {"cmd.run_all": MagicMock(return_value={"retcode": 0})}
        ):
            expected = {}
            result = win_pkg.install(name="nsis", version="3.03")
            self.assertDictEqual(expected, result)

    def test_pkg_install_name(self):
        """
        test pkg.install name extra_install_flags
        """

        ret__get_package_info = {
            "3.03": {
                "uninstaller": "%program.exe",
                "reboot": False,
                "msiexec": False,
                "installer": "runme.exe",
                "uninstall_flags": "/S",
                "locale": "en_US",
                "install_flags": "/s",
                "full_name": "Firebox 3.03 (x86 en-US)",
            }
        }

        mock_cmd_run_all = MagicMock(return_value={"retcode": 0})
        with patch.object(
            salt.utils.data, "is_true", MagicMock(return_value=True)
        ), patch.object(
            win_pkg, "_get_package_info", MagicMock(return_value=ret__get_package_info)
        ), patch.dict(
            win_pkg.__salt__,
            {
                "pkg_resource.parse_targets": MagicMock(
                    return_value=[{"firebox": "3.03"}, None]
                ),
                "cp.is_cached": MagicMock(return_value="C:\\fake\\path.exe"),
                "cmd.run_all": mock_cmd_run_all,
            },
        ):
            ret = win_pkg.install(
                name="firebox",
                version="3.03",
                extra_install_flags="-e True -test_flag True",
            )
            self.assertTrue(
                "-e True -test_flag True" in str(mock_cmd_run_all.call_args[0])
            )

    def test_pkg_install_single_pkg(self):
        """
        test pkg.install pkg with extra_install_flags
        """
        ret__get_package_info = {
            "3.03": {
                "uninstaller": "%program.exe",
                "reboot": False,
                "msiexec": False,
                "installer": "runme.exe",
                "uninstall_flags": "/S",
                "locale": "en_US",
                "install_flags": "/s",
                "full_name": "Firebox 3.03 (x86 en-US)",
            }
        }

        mock_cmd_run_all = MagicMock(return_value={"retcode": 0})
        with patch.object(
            salt.utils.data, "is_true", MagicMock(return_value=True)
        ), patch.object(
            win_pkg, "_get_package_info", MagicMock(return_value=ret__get_package_info)
        ), patch.dict(
            win_pkg.__salt__,
            {
                "pkg_resource.parse_targets": MagicMock(
                    return_value=[{"firebox": "3.03"}, None]
                ),
                "cp.is_cached": MagicMock(return_value="C:\\fake\\path.exe"),
                "cmd.run_all": mock_cmd_run_all,
            },
        ):
            ret = win_pkg.install(
                pkgs=["firebox"],
                version="3.03",
                extra_install_flags="-e True -test_flag True",
            )
            self.assertTrue(
                "-e True -test_flag True" in str(mock_cmd_run_all.call_args[0])
            )

    def test_pkg_install_multiple_pkgs(self):
        """
        test pkg.install pkg with extra_install_flags
        """
        ret__get_package_info = {
            "3.03": {
                "uninstaller": "%program.exe",
                "reboot": False,
                "msiexec": False,
                "installer": "runme.exe",
                "uninstall_flags": "/S",
                "locale": "en_US",
                "install_flags": "/s",
                "full_name": "Firebox 3.03 (x86 en-US)",
            }
        }

        mock_cmd_run_all = MagicMock(return_value={"retcode": 0})
        with patch.object(
            salt.utils.data, "is_true", MagicMock(return_value=True)
        ), patch.object(
            win_pkg, "_get_package_info", MagicMock(return_value=ret__get_package_info)
        ), patch.dict(
            win_pkg.__salt__,
            {
                "pkg_resource.parse_targets": MagicMock(
                    return_value=[{"firebox": "3.03", "got": "3.03"}, None]
                ),
                "cp.is_cached": MagicMock(return_value="C:\\fake\\path.exe"),
                "cmd.run_all": mock_cmd_run_all,
            },
        ):
            ret = win_pkg.install(
                pkgs=["firebox", "got"], extra_install_flags="-e True -test_flag True"
            )
            self.assertFalse(
                "-e True -test_flag True" in str(mock_cmd_run_all.call_args[0])
            )

    def test_pkg_install_minion_error_https(self):
        """
        Test pkg.install when cp.cache_file encounters a minion error
        """
        ret__get_package_info = {
            "3.03": {
                "uninstaller": "%program.exe",
                "reboot": False,
                "msiexec": False,
                "installer": "https://repo.test.com/runme.exe",
                "uninstall_flags": "/S",
                "locale": "en_US",
                "install_flags": "/s",
                "full_name": "Firebox 3.03 (x86 en-US)",
            }
        }

        err_msg = (
            "Error: [Errno 11001] getaddrinfo failed reading"
            " https://repo.test.com/runme.exe"
        )
        mock_none = MagicMock(return_value=None)
        mock_minion_error = MagicMock(side_effect=MinionError(err_msg))
        mock_parse = MagicMock(return_value=[{"firebox": "3.03"}, None])
        with patch.object(
            salt.utils.data, "is_true", MagicMock(return_value=True)
        ), patch.object(
            win_pkg, "_get_package_info", MagicMock(return_value=ret__get_package_info)
        ), patch.dict(
            win_pkg.__salt__,
            {
                "pkg_resource.parse_targets": mock_parse,
                "cp.is_cached": mock_none,
                "cp.cache_file": mock_minion_error,
            },
        ):
            ret = win_pkg.install(
                name="firebox",
                version="3.03",
            )
            expected = (
                "Failed to cache https://repo.test.com/runme.exe\nError: [Errno 11001]"
                " getaddrinfo failed reading https://repo.test.com/runme.exe"
            )

            self.assertEqual(ret, expected)

    def test_pkg_install_minion_error_salt(self):
        """
        Test pkg.install when cp.cache_file encounters a minion error
        """
        ret__get_package_info = {
            "3.03": {
                "uninstaller": "%program.exe",
                "reboot": False,
                "msiexec": False,
                "installer": "salt://software/runme.exe",
                "uninstall_flags": "/S",
                "locale": "en_US",
                "install_flags": "/s",
                "full_name": "Firebox 3.03 (x86 en-US)",
            }
        }

        err_msg = "Error: [Errno 1] failed reading salt://software/runme.exe"
        mock_none = MagicMock(return_value=None)
        mock_minion_error = MagicMock(side_effect=MinionError(err_msg))
        mock_parse = MagicMock(return_value=[{"firebox": "3.03"}, None])
        with patch.object(
            salt.utils.data, "is_true", MagicMock(return_value=True)
        ), patch.object(
            win_pkg, "_get_package_info", MagicMock(return_value=ret__get_package_info)
        ), patch.dict(
            win_pkg.__salt__,
            {
                "pkg_resource.parse_targets": mock_parse,
                "cp.is_cached": mock_none,
                "cp.cache_file": mock_minion_error,
            },
        ):
            ret = win_pkg.install(
                name="firebox",
                version="3.03",
            )
            expected = (
                "Failed to cache salt://software/runme.exe\n"
                "Error: [Errno 1] failed reading salt://software/runme.exe"
            )

            self.assertEqual(ret, expected)

    def test_pkg_install_minion_error_salt_cache_dir(self):
        """
        Test pkg.install when cp.cache_dir encounters a minion error
        """
        ret__get_package_info = {
            "3.03": {
                "uninstaller": "%program.exe",
                "reboot": False,
                "msiexec": False,
                "installer": "salt://software/runme.exe",
                "cache_dir": True,
                "uninstall_flags": "/S",
                "locale": "en_US",
                "install_flags": "/s",
                "full_name": "Firebox 3.03 (x86 en-US)",
            }
        }

        err_msg = "Error: [Errno 1] failed reading salt://software"
        mock_none = MagicMock(return_value=None)
        mock_minion_error = MagicMock(side_effect=MinionError(err_msg))
        mock_parse = MagicMock(return_value=[{"firebox": "3.03"}, None])
        with patch.object(
            salt.utils.data, "is_true", MagicMock(return_value=True)
        ), patch.object(
            win_pkg, "_get_package_info", MagicMock(return_value=ret__get_package_info)
        ), patch.dict(
            win_pkg.__salt__,
            {"cp.cache_dir": mock_minion_error},
        ):
            ret = win_pkg.install(
                name="firebox",
                version="3.03",
            )
            expected = (
                "Failed to cache salt://software\n"
                "Error: [Errno 1] failed reading salt://software"
            )

            self.assertEqual(ret, expected)

    def test_pkg_remove_minion_error_salt_cache_dir(self):
        """
        Test pkg.remove when cp.cache_dir encounters a minion error
        """
        ret__get_package_info = {
            "3.03": {
                "uninstaller": "salt://software/runme.exe",
                "reboot": False,
                "msiexec": False,
                "installer": "salt://software/runme.exe",
                "cache_dir": True,
                "uninstall_flags": "/U /S",
                "locale": "en_US",
                "install_flags": "/s",
                "full_name": "Firebox 3.03 (x86 en-US)",
            }
        }

        err_msg = "Error: [Errno 1] failed reading salt://software"
        mock_minion_error = MagicMock(side_effect=MinionError(err_msg))
        mock_parse = MagicMock(return_value=[{"firebox": "3.03"}, None])
        se_list_pkgs = {"firebox": ["3.03"]}
        with patch.object(
            win_pkg, "list_pkgs", return_value=se_list_pkgs
        ), patch.object(
            salt.utils.data, "is_true", MagicMock(return_value=True)
        ), patch.object(
            win_pkg, "_get_package_info", MagicMock(return_value=ret__get_package_info)
        ), patch.dict(
            win_pkg.__salt__,
            {
                "pkg_resource.parse_targets": mock_parse,
                "cp.cache_dir": mock_minion_error,
            },
        ):
            ret = win_pkg.remove(name="firebox")
            expected = (
                "Failed to cache salt://software\n"
                "Error: [Errno 1] failed reading salt://software"
            )

            self.assertEqual(ret, expected)

    def test_pkg_remove_minion_error_salt(self):
        """
        Test pkg.remove when cp.cache_file encounters a minion error
        """
        ret__get_package_info = {
            "3.03": {
                "uninstaller": "salt://software/runme.exe",
                "reboot": False,
                "msiexec": False,
                "installer": "salt://software/runme.exe",
                "uninstall_flags": "/U /S",
                "locale": "en_US",
                "install_flags": "/s",
                "full_name": "Firebox 3.03 (x86 en-US)",
            }
        }

        err_msg = "Error: [Errno 1] failed reading salt://software/runme.exe"
        mock_minion_error = MagicMock(side_effect=MinionError(err_msg))
        mock_none = MagicMock(return_value=None)
        mock_parse = MagicMock(return_value=[{"firebox": "3.03"}, None])
        se_list_pkgs = {"firebox": ["3.03"]}
        with patch.object(
            win_pkg, "list_pkgs", return_value=se_list_pkgs
        ), patch.object(
            salt.utils.data, "is_true", MagicMock(return_value=True)
        ), patch.object(
            win_pkg, "_get_package_info", MagicMock(return_value=ret__get_package_info)
        ), patch.dict(
            win_pkg.__salt__,
            {
                "pkg_resource.parse_targets": mock_parse,
                "cp.is_cached": mock_none,
                "cp.cache_file": mock_minion_error,
            },
        ):
            ret = win_pkg.remove(name="firebox")
            expected = (
                "Failed to cache salt://software/runme.exe\n"
                "Error: [Errno 1] failed reading salt://software/runme.exe"
            )

            self.assertEqual(ret, expected)



"""
Integration tests for the etcd modules
"""

import logging

import pytest

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.skip_if_binaries_missing("dockerd"),
]


@pytest.fixture(scope="module", autouse=True)
def etc_docker_container(salt_call_cli, sdb_etcd_port):
    container_started = False
    try:
        ret = salt_call_cli.run(
            "state.single", "docker_image.present", name="bitnami/etcd", tag="latest"
        )
        assert ret.exitcode == 0
        assert ret.json
        state_run = next(iter(ret.json.values()))
        assert state_run["result"] is True
        ret = salt_call_cli.run(
            "state.single",
            "docker_container.running",
            name="etcd",
            image="bitnami/etcd:latest",
            port_bindings="{}:2379".format(sdb_etcd_port),
            environment={"ALLOW_NONE_AUTHENTICATION": "yes", "ETCD_ENABLE_V2": "true"},
            cap_add="IPC_LOCK",
        )
        assert ret.exitcode == 0
        assert ret.json
        state_run = next(iter(ret.json.values()))
        assert state_run["result"] is True
        container_started = True
        yield
    finally:
        if container_started:
            ret = salt_call_cli.run(
                "state.single", "docker_container.stopped", name="etcd"
            )
            assert ret.exitcode == 0
            assert ret.json
            state_run = next(iter(ret.json.values()))
            assert state_run["result"] is True
            ret = salt_call_cli.run(
                "state.single", "docker_container.absent", name="etcd"
            )
            assert ret.exitcode == 0
            assert ret.json
            state_run = next(iter(ret.json.values()))
            assert state_run["result"] is True


@pytest.fixture(scope="module")
def pillar_tree(base_env_pillar_tree_root_dir, salt_minion):
    top_file = """
    base:
      '{}':
        - sdb
    """.format(
        salt_minion.id
    )
    sdb_pillar_file = """
    test_vault_pillar_sdb: sdb://sdbvault/secret/test/test_pillar_sdb/foo
    test_etcd_pillar_sdb: sdb://sdbetcd/secret/test/test_pillar_sdb/foo
    """
    top_tempfile = pytest.helpers.temp_file(
        "top.sls", top_file, base_env_pillar_tree_root_dir
    )
    sdb_tempfile = pytest.helpers.temp_file(
        "sdb.sls", sdb_pillar_file, base_env_pillar_tree_root_dir
    )

    with top_tempfile, sdb_tempfile:
        yield


@pytest.mark.slow_test
def test_sdb(salt_call_cli):
    ret = salt_call_cli.run(
        "sdb.set", uri="sdb://sdbetcd/secret/test/test_sdb/foo", value="bar"
    )
    assert ret.exitcode == 0
    assert ret.json
    assert ret.json == "bar"

    ret = salt_call_cli.run("sdb.get", uri="sdb://sdbetcd/secret/test/test_sdb/foo")
    assert ret.exitcode == 0
    assert ret.json
    assert ret.json == "bar"


@pytest.mark.slow_test
def test_sdb_runner(salt_run_cli):
    ret = salt_run_cli.run(
        "sdb.set", uri="sdb://sdbetcd/secret/test/test_sdb_runner/foo", value="bar"
    )
    assert ret.exitcode == 0
    assert ret.stdout == "bar"

    ret = salt_run_cli.run(
        "sdb.get", uri="sdb://sdbetcd/secret/test/test_sdb_runner/foo"
    )
    assert ret.exitcode == 0
    assert ret.stdout == "bar"


@pytest.mark.slow_test
def test_config(salt_call_cli, pillar_tree):
    ret = salt_call_cli.run(
        "sdb.set", uri="sdb://sdbetcd/secret/test/test_pillar_sdb/foo", value="bar"
    )
    assert ret.exitcode == 0
    assert ret.json
    assert ret.json == "bar"

    ret = salt_call_cli.run("config.get", "test_etcd_pillar_sdb")
    assert ret.exitcode == 0
    assert ret.json
    assert ret.json == "bar"



"""
Alex Martelli's soulution for recursive dict update from
http://stackoverflow.com/a/3233356
"""

import copy
import logging
from collections.abc import Mapping

import salt.utils.data
from salt.defaults import DEFAULT_TARGET_DELIM
from salt.exceptions import SaltInvocationError
from salt.utils.decorators.jinja import jinja_filter
from salt.utils.odict import OrderedDict

log = logging.getLogger(__name__)


def update(dest, upd, recursive_update=True, merge_lists=False):
    """
    Recursive version of the default dict.update

    Merges upd recursively into dest

    If recursive_update=False, will use the classic dict.update, or fall back
    on a manual merge (helpful for non-dict types like FunctionWrapper)

    If merge_lists=True, will aggregate list object types instead of replace.
    The list in ``upd`` is added to the list in ``dest``, so the resulting list
    is ``dest[key] + upd[key]``. This behavior is only activated when
    recursive_update=True. By default merge_lists=False.

    .. versionchanged:: 2016.11.6
        When merging lists, duplicate values are removed. Values already
        present in the ``dest`` list are not added from the ``upd`` list.
    """
    if (not isinstance(dest, Mapping)) or (not isinstance(upd, Mapping)):
        raise TypeError("Cannot update using non-dict types in dictupdate.update()")
    updkeys = list(upd.keys())
    if not set(list(dest.keys())) & set(updkeys):
        recursive_update = False
    if recursive_update:
        for key in updkeys:
            val = upd[key]
            try:
                dest_subkey = dest.get(key, None)
            except AttributeError:
                dest_subkey = None
            if isinstance(dest_subkey, Mapping) and isinstance(val, Mapping):
                ret = update(dest_subkey, val, merge_lists=merge_lists)
                dest[key] = ret
            elif isinstance(dest_subkey, list) and isinstance(val, list):
                if merge_lists:
                    merged = copy.deepcopy(dest_subkey)
                    merged.extend([x for x in val if x not in merged])
                    dest[key] = merged
                else:
                    dest[key] = upd[key]
            else:
                dest[key] = upd[key]
        return dest
    for k in upd:
        dest[k] = upd[k]
    return dest


def merge_list(obj_a, obj_b):
    ret = {}
    for key, val in obj_a.items():
        if key in obj_b:
            ret[key] = [val, obj_b[key]]
        else:
            ret[key] = val
    return ret


def merge_recurse(obj_a, obj_b, merge_lists=False):
    copied = copy.deepcopy(obj_a)
    return update(copied, obj_b, merge_lists=merge_lists)


def merge_aggregate(obj_a, obj_b):
    from salt.serializers.yamlex import merge_recursive as _yamlex_merge_recursive

    return _yamlex_merge_recursive(obj_a, obj_b, level=1)


def merge_overwrite(obj_a, obj_b, merge_lists=False):
    for obj in obj_b:
        if obj in obj_a:
            obj_a[obj] = obj_b[obj]
    return merge_recurse(obj_a, obj_b, merge_lists=merge_lists)


def merge(obj_a, obj_b, strategy="smart", renderer="yaml", merge_lists=False):
    if strategy == "smart":
        if renderer.split("|")[-1] == "yamlex" or renderer.startswith("yamlex_"):
            strategy = "aggregate"
        else:
            strategy = "recurse"

    if strategy == "list":
        merged = merge_list(obj_a, obj_b)
    elif strategy == "recurse":
        merged = merge_recurse(obj_a, obj_b, merge_lists)
    elif strategy == "aggregate":
        #: level = 1 merge at least root data
        merged = merge_aggregate(obj_a, obj_b)
    elif strategy == "overwrite":
        merged = merge_overwrite(obj_a, obj_b, merge_lists)
    elif strategy == "none":
        # If we do not want to merge, there is only one pillar passed, so we can safely use the default recurse,
        # we just do not want to log an error
        merged = merge_recurse(obj_a, obj_b)
    else:
        log.warning("Unknown merging strategy '%s', fallback to recurse", strategy)
        merged = merge_recurse(obj_a, obj_b)

    return merged


def ensure_dict_key(in_dict, keys, delimiter=DEFAULT_TARGET_DELIM, ordered_dict=False):
    """
    Ensures that in_dict contains the series of recursive keys defined in keys.

    :param dict in_dict: The dict to work with.
    :param str keys: The delimited string with one or more keys.
    :param str delimiter: The delimiter to use in `keys`. Defaults to ':'.
    :param bool ordered_dict: Create OrderedDicts if keys are missing.
                              Default: create regular dicts.
    :rtype: dict
    :return: Returns the modified in-place `in_dict`.
    """
    if delimiter in keys:
        a_keys = keys.split(delimiter)
    else:
        a_keys = [keys]
    dict_pointer = in_dict
    while a_keys:
        current_key = a_keys.pop(0)
        if current_key not in dict_pointer or not isinstance(
            dict_pointer[current_key], dict
        ):
            dict_pointer[current_key] = OrderedDict() if ordered_dict else {}
        dict_pointer = dict_pointer[current_key]
    return in_dict


def _dict_rpartition(in_dict, keys, delimiter=DEFAULT_TARGET_DELIM, ordered_dict=False):
    """
    Helper function to:
    - Ensure all but the last key in `keys` exist recursively in `in_dict`.
    - Return the dict at the one-to-last key, and the last key

    :param dict in_dict: The dict to work with.
    :param str keys: The delimited string with one or more keys.
    :param str delimiter: The delimiter to use in `keys`. Defaults to ':'.
    :param bool ordered_dict: Create OrderedDicts if keys are missing.
                              Default: create regular dicts.
    :rtype: tuple(dict, str)
    :return: (The dict at the one-to-last key, the last key)
    """
    if delimiter in keys:
        all_but_last_keys, _, last_key = keys.rpartition(delimiter)
        ensure_dict_key(
            in_dict, all_but_last_keys, delimiter=delimiter, ordered_dict=ordered_dict
        )
        dict_pointer = salt.utils.data.traverse_dict(
            in_dict, all_but_last_keys, default=None, delimiter=delimiter
        )
    else:
        dict_pointer = in_dict
        last_key = keys
    return dict_pointer, last_key


@jinja_filter("set_dict_key_value")
def set_dict_key_value(
    in_dict, keys, value, delimiter=DEFAULT_TARGET_DELIM, ordered_dict=False
):
    """
    Ensures that in_dict contains the series of recursive keys defined in keys.
    Also sets whatever is at the end of `in_dict` traversed with `keys` to `value`.

    :param dict in_dict: The dictionary to work with
    :param str keys: The delimited string with one or more keys.
    :param any value: The value to assign to the nested dict-key.
    :param str delimiter: The delimiter to use in `keys`. Defaults to ':'.
    :param bool ordered_dict: Create OrderedDicts if keys are missing.
                              Default: create regular dicts.
    :rtype: dict
    :return: Returns the modified in-place `in_dict`.
    """
    dict_pointer, last_key = _dict_rpartition(
        in_dict, keys, delimiter=delimiter, ordered_dict=ordered_dict
    )
    dict_pointer[last_key] = value
    return in_dict


@jinja_filter("update_dict_key_value")
def update_dict_key_value(
    in_dict, keys, value, delimiter=DEFAULT_TARGET_DELIM, ordered_dict=False
):
    """
    Ensures that in_dict contains the series of recursive keys defined in keys.
    Also updates the dict, that is at the end of `in_dict` traversed with `keys`,
    with `value`.

    :param dict in_dict: The dictionary to work with
    :param str keys: The delimited string with one or more keys.
    :param any value: The value to update the nested dict-key with.
    :param str delimiter: The delimiter to use in `keys`. Defaults to ':'.
    :param bool ordered_dict: Create OrderedDicts if keys are missing.
                              Default: create regular dicts.
    :rtype: dict
    :return: Returns the modified in-place `in_dict`.
    """
    dict_pointer, last_key = _dict_rpartition(
        in_dict, keys, delimiter=delimiter, ordered_dict=ordered_dict
    )
    if last_key not in dict_pointer or dict_pointer[last_key] is None:
        dict_pointer[last_key] = OrderedDict() if ordered_dict else {}
    try:
        dict_pointer[last_key].update(value)
    except AttributeError:
        raise SaltInvocationError(
            "The last key contains a {}, which cannot update.".format(
                type(dict_pointer[last_key])
            )
        )
    except (ValueError, TypeError):
        raise SaltInvocationError(
            "Cannot update {} with a {}.".format(
                type(dict_pointer[last_key]), type(value)
            )
        )
    return in_dict


@jinja_filter("append_dict_key_value")
def append_dict_key_value(
    in_dict, keys, value, delimiter=DEFAULT_TARGET_DELIM, ordered_dict=False
):
    """
    Ensures that in_dict contains the series of recursive keys defined in keys.
    Also appends `value` to the list that is at the end of `in_dict` traversed
    with `keys`.

    :param dict in_dict: The dictionary to work with
    :param str keys: The delimited string with one or more keys.
    :param any value: The value to append to the nested dict-key.
    :param str delimiter: The delimiter to use in `keys`. Defaults to ':'.
    :param bool ordered_dict: Create OrderedDicts if keys are missing.
                              Default: create regular dicts.
    :rtype: dict
    :return: Returns the modified in-place `in_dict`.
    """
    dict_pointer, last_key = _dict_rpartition(
        in_dict, keys, delimiter=delimiter, ordered_dict=ordered_dict
    )
    if last_key not in dict_pointer or dict_pointer[last_key] is None:
        dict_pointer[last_key] = []
    try:
        dict_pointer[last_key].append(value)
    except AttributeError:
        raise SaltInvocationError(
            "The last key contains a {}, which cannot append.".format(
                type(dict_pointer[last_key])
            )
        )
    return in_dict


@jinja_filter("extend_dict_key_value")
def extend_dict_key_value(
    in_dict, keys, value, delimiter=DEFAULT_TARGET_DELIM, ordered_dict=False
):
    """
    Ensures that in_dict contains the series of recursive keys defined in keys.
    Also extends the list, that is at the end of `in_dict` traversed with `keys`,
    with `value`.

    :param dict in_dict: The dictionary to work with
    :param str keys: The delimited string with one or more keys.
    :param any value: The value to extend the nested dict-key with.
    :param str delimiter: The delimiter to use in `keys`. Defaults to ':'.
    :param bool ordered_dict: Create OrderedDicts if keys are missing.
                              Default: create regular dicts.
    :rtype: dict
    :return: Returns the modified in-place `in_dict`.
    """
    dict_pointer, last_key = _dict_rpartition(
        in_dict, keys, delimiter=delimiter, ordered_dict=ordered_dict
    )
    if last_key not in dict_pointer or dict_pointer[last_key] is None:
        dict_pointer[last_key] = []
    try:
        dict_pointer[last_key].extend(value)
    except AttributeError:
        raise SaltInvocationError(
            "The last key contains a {}, which cannot extend.".format(
                type(dict_pointer[last_key])
            )
        )
    except TypeError:
        raise SaltInvocationError(
            "Cannot extend {} with a {}.".format(
                type(dict_pointer[last_key]), type(value)
            )
        )
    return in_dict



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import urllib.error

import salt.modules.apache as apache
from salt.utils.odict import OrderedDict
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, mock_open, patch
from tests.support.unit import TestCase


class ApacheTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.apache
    """

    def setup_loader_modules(self):
        return {apache: {}}

    # 'version' function tests: 1
    def test_version(self):
        """
        Test if return server version (``apachectl -v``)
        """
        with patch(
            "salt.modules.apache._detect_os", MagicMock(return_value="apachectl")
        ):
            mock = MagicMock(return_value="Server version: Apache/2.4.7")
            with patch.dict(apache.__salt__, {"cmd.run": mock}):
                assert apache.version() == "Apache/2.4.7"

    # 'fullversion' function tests: 1

    def test_fullversion(self):
        """
        Test if return server version (``apachectl -V``)
        """
        with patch(
            "salt.modules.apache._detect_os", MagicMock(return_value="apachectl")
        ):
            mock = MagicMock(return_value="Server version: Apache/2.4.7")
            with patch.dict(apache.__salt__, {"cmd.run": mock}):
                assert apache.fullversion() == {
                    "compiled_with": [],
                    "server_version": "Apache/2.4.7",
                }

    # 'modules' function tests: 1

    def test_modules(self):
        """
        Test if return list of static and shared modules
        """
        with patch(
            "salt.modules.apache._detect_os", MagicMock(return_value="apachectl")
        ):
            mock = MagicMock(
                return_value=(
                    "unixd_module (static)\n                             "
                    " access_compat_module (shared)"
                )
            )
            with patch.dict(apache.__salt__, {"cmd.run": mock}):
                assert apache.modules() == {
                    "shared": ["access_compat_module"],
                    "static": ["unixd_module"],
                }

    # 'servermods' function tests: 1

    def test_servermods(self):
        """
        Test if return list of modules compiled into the server
        """
        with patch(
            "salt.modules.apache._detect_os", MagicMock(return_value="apachectl")
        ):
            mock = MagicMock(return_value="core.c\nmod_so.c")
            with patch.dict(apache.__salt__, {"cmd.run": mock}):
                assert apache.servermods() == ["core.c", "mod_so.c"]

    # 'directives' function tests: 1

    def test_directives(self):
        """
        Test if return list of directives
        """
        with patch(
            "salt.modules.apache._detect_os", MagicMock(return_value="apachectl")
        ):
            mock = MagicMock(return_value="Salt")
            with patch.dict(apache.__salt__, {"cmd.run": mock}):
                assert apache.directives() == {"Salt": ""}

    # 'vhosts' function tests: 1

    def test_vhosts(self):
        """
        Test if it shows the virtualhost settings
        """
        with patch(
            "salt.modules.apache._detect_os", MagicMock(return_value="apachectl")
        ):
            mock = MagicMock(return_value="")
            with patch.dict(apache.__salt__, {"cmd.run": mock}):
                assert apache.vhosts() == {}

    # 'signal' function tests: 2

    def test_signal(self):
        """
        Test if return no signal for httpd
        """
        with patch(
            "salt.modules.apache._detect_os", MagicMock(return_value="apachectl")
        ):
            mock = MagicMock(return_value="")
            with patch.dict(apache.__salt__, {"cmd.run": mock}):
                assert apache.signal(None) is None

    def test_signal_args(self):
        """
        Test if return httpd signal to start, restart, or stop.
        """
        with patch(
            "salt.modules.apache._detect_os", MagicMock(return_value="apachectl")
        ):
            ret = 'Command: "apachectl -k start" completed successfully!'
            mock = MagicMock(return_value={"retcode": 1, "stderr": "", "stdout": ""})
            with patch.dict(apache.__salt__, {"cmd.run_all": mock}):
                assert apache.signal("start") == ret

            mock = MagicMock(
                return_value={"retcode": 1, "stderr": "Syntax OK", "stdout": ""}
            )
            with patch.dict(apache.__salt__, {"cmd.run_all": mock}):
                assert apache.signal("start") == "Syntax OK"

            mock = MagicMock(
                return_value={"retcode": 0, "stderr": "Syntax OK", "stdout": ""}
            )
            with patch.dict(apache.__salt__, {"cmd.run_all": mock}):
                assert apache.signal("start") == "Syntax OK"

            mock = MagicMock(
                return_value={"retcode": 1, "stderr": "", "stdout": "Salt"}
            )
            with patch.dict(apache.__salt__, {"cmd.run_all": mock}):
                assert apache.signal("start") == "Salt"

    # 'useradd' function tests: 1

    def test_useradd(self):
        """
        Test if it add HTTP user using the ``htpasswd`` command
        """
        mock = MagicMock(return_value=True)
        with patch.dict(apache.__salt__, {"webutil.useradd": mock}):
            assert apache.useradd("htpasswd", "salt", "badpassword") is True

    # 'userdel' function tests: 1

    def test_userdel(self):
        """
        Test if it delete HTTP user using the ``htpasswd`` file
        """
        mock = MagicMock(return_value=True)
        with patch.dict(apache.__salt__, {"webutil.userdel": mock}):
            assert apache.userdel("htpasswd", "salt") is True

    # 'server_status' function tests: 2

    def test_server_status(self):
        """
        Test if return get information from the Apache server-status
        """
        with patch("salt.modules.apache.server_status", MagicMock(return_value={})):
            mock = MagicMock(return_value="")
            with patch.dict(apache.__salt__, {"config.get": mock}):
                assert apache.server_status() == {}

    def test_server_status_error(self):
        """
        Test if return get error from the Apache server-status
        """
        mock = MagicMock(side_effect=urllib.error.URLError("error"))
        with patch("urllib.request.urlopen", mock):
            mock = MagicMock(return_value="")
            with patch.dict(apache.__salt__, {"config.get": mock}):
                assert apache.server_status() == "error"

    # 'config' function tests: 1

    def test_config(self):
        """
        Test if it create VirtualHost configuration files
        """
        with patch(
            "salt.modules.apache._parse_config", MagicMock(return_value="Listen 22")
        ):
            with patch("salt.utils.files.fopen", mock_open()):
                assert apache.config("/ports.conf", [{"Listen": "22"}]) == "Listen 22"

    # '_parse_config' function tests: 2

    def test__parse_config_dict(self):
        """
        Test parsing function which creates configs from dict like (legacy way):
            - VirtualHost:
              this: '*:80'
              ServerName: website.com
              ServerAlias:
                - www
                - dev
              Directory:
                  this: /var/www/vhosts/website.com
                  Order: Deny,Allow
                  Allow from:
                    - 127.0.0.1
                    - 192.168.100.0/24

        """
        data_in = OrderedDict(
            [
                (
                    "Directory",
                    OrderedDict(
                        [
                            ("this", "/var/www/vhosts/website.com"),
                            ("Order", "Deny,Allow"),
                            ("Allow from", ["127.0.0.1", "192.168.100.0/24"]),
                        ]
                    ),
                ),
                ("this", "*:80"),
                ("ServerName", "website.com"),
                ("ServerAlias", ["www", "dev"]),
            ]
        )
        dataout = (
            "<VirtualHost *:80>\n"
            "<Directory /var/www/vhosts/website.com>\n"
            "Order Deny,Allow\n"
            "Allow from 127.0.0.1\n"
            "Allow from 192.168.100.0/24\n\n"
            "</Directory>\n\n"
            "ServerName website.com\n"
            "ServerAlias www\n"
            "ServerAlias dev\n\n"
            "</VirtualHost>\n"
        )
        # pylint: disable=protected-access
        parse = apache._parse_config(data_in, "VirtualHost")
        assert parse == dataout

    def test__parse_config_list(self):
        """
        Test parsing function which creates configs from variable structure (list of dicts or
        list of dicts of dicts/lists) like:
            - VirtualHost:
              - this: '*:80'
              - ServerName: website.com
              - ServerAlias:
                - www
                - dev
              - Directory:
                  this: /var/www/vhosts/website.com
                  Order: Deny,Allow
                  Allow from:
                    - 127.0.0.1
                    - 192.168.100.0/24
              - Directory:
                - this: /var/www/vhosts/website.com/private
                - Order: Deny,Allow
                - Allow from:
                  - 127.0.0.1
                  - 192.168.100.0/24
                - If:
                    this: some condition
                    do: something
        """
        data_in = [
            OrderedDict(
                [
                    ("ServerName", "website.com"),
                    ("ServerAlias", ["www", "dev"]),
                    (
                        "Directory",
                        [
                            OrderedDict(
                                [
                                    ("this", "/var/www/vhosts/website.com/private"),
                                    ("Order", "Deny,Allow"),
                                    ("Allow from", ["127.0.0.1", "192.168.100.0/24"]),
                                    (
                                        "If",
                                        {"this": "some condition", "do": "something"},
                                    ),
                                ]
                            )
                        ],
                    ),
                    ("this", "*:80"),
                ]
            )
        ]
        dataout = (
            "<VirtualHost *:80>\n"
            "ServerName website.com\n"
            "ServerAlias www\n"
            "ServerAlias dev\n\n"
            "<Directory /var/www/vhosts/website.com/private>\n"
            "Order Deny,Allow\n"
            "Allow from 127.0.0.1\n"
            "Allow from 192.168.100.0/24\n\n"
            "<If some condition>\n"
            "do something\n"
            "</If>\n\n"
            "</Directory>\n\n"
            "</VirtualHost>\n"
        )
        # pylint: disable=protected-access
        parse = apache._parse_config(data_in, "VirtualHost")
        assert parse == dataout



"""
    This is a simple proxy-minion designed to connect to and communicate with
    a server that exposes functionality via SSH.
    This can be used as an option when the device does not provide
    an api over HTTP and doesn't have the python stack to run a minion.
"""

import logging

import salt.utils.json
from salt.utils.vt import TerminalException
from salt.utils.vt_helper import SSHConnection

# This must be present or the Salt loader won't load this module
__proxyenabled__ = ["ssh_sample"]

DETAILS = {}

log = logging.getLogger(__file__)


# This does nothing, it's here just as an example and to provide a log
# entry when the module is loaded.
def __virtual__():
    """
    Only return if all the modules are available
    """
    log.info("ssh_sample proxy __virtual__() called...")

    return True


def init(opts):
    """
    Required.
    Can be used to initialize the server connection.
    """
    try:
        DETAILS["server"] = SSHConnection(
            host=__opts__["proxy"]["host"],
            username=__opts__["proxy"]["username"],
            password=__opts__["proxy"]["password"],
        )
        out, err = DETAILS["server"].sendline("help")
        DETAILS["initialized"] = True

    except TerminalException as e:
        log.error(e)
        return False


def initialized():
    """
    Since grains are loaded in many different places and some of those
    places occur before the proxy can be initialized, return whether
    our init() function has been called
    """
    return DETAILS.get("initialized", False)


def grains():
    """
    Get the grains from the proxied device
    """

    if not DETAILS.get("grains_cache", {}):
        cmd = "info"

        # Send the command to execute
        out, err = DETAILS["server"].sendline(cmd)

        # "scrape" the output and return the right fields as a dict
        DETAILS["grains_cache"] = parse(out)

    return DETAILS["grains_cache"]


def grains_refresh():
    """
    Refresh the grains from the proxied device
    """
    DETAILS["grains_cache"] = None
    return grains()


def fns():
    """
    Method called by grains module.
    """
    return {
        "details": (
            "This key is here because a function in "
            "grains/ssh_sample.py called fns() here in the proxymodule."
        )
    }


def ping():
    """
    Required.
    Ping the device on the other end of the connection
    """
    try:
        out, err = DETAILS["server"].sendline("help")
        return True
    except TerminalException as e:
        log.error(e)
        return False


def shutdown(opts):
    """
    Disconnect
    """
    DETAILS["server"].close_connection()


def parse(out):
    """
    Extract json from out.

    Parameter
        out: Type string. The data returned by the
        ssh command.
    """
    jsonret = []
    in_json = False
    for ln_ in out.split("\n"):
        if "{" in ln_:
            in_json = True
        if in_json:
            jsonret.append(ln_)
        if "}" in ln_:
            in_json = False
    return salt.utils.json.loads("\n".join(jsonret))


def package_list():
    """
    List "packages" by executing a command via ssh
    This function is called in response to the salt command

    .. code-block:: bash

        salt target_minion pkg.list_pkgs

    """
    # Send the command to execute
    out, err = DETAILS["server"].sendline("pkg_list\n")

    # "scrape" the output and return the right fields as a dict
    return parse(out)


def package_install(name, **kwargs):
    """
    Install a "package" on the ssh server
    """
    cmd = "pkg_install " + name
    if kwargs.get("version", False):
        cmd += " " + kwargs["version"]

    # Send the command to execute
    out, err = DETAILS["server"].sendline(cmd)

    # "scrape" the output and return the right fields as a dict
    return parse(out)


def package_remove(name):
    """
    Remove a "package" on the ssh server
    """
    cmd = "pkg_remove " + name

    # Send the command to execute
    out, err = DETAILS["server"].sendline(cmd)

    # "scrape" the output and return the right fields as a dict
    return parse(out)


def service_list():
    """
    Start a "service" on the ssh server

    .. versionadded:: 2015.8.2
    """
    cmd = "ps"

    # Send the command to execute
    out, err = DETAILS["server"].sendline(cmd)

    # "scrape" the output and return the right fields as a dict
    return parse(out)


def service_start(name):
    """
    Start a "service" on the ssh server

    .. versionadded:: 2015.8.2
    """
    cmd = "start " + name

    # Send the command to execute
    out, err = DETAILS["server"].sendline(cmd)

    # "scrape" the output and return the right fields as a dict
    return parse(out)


def service_stop(name):
    """
    Stop a "service" on the ssh server

    .. versionadded:: 2015.8.2
    """
    cmd = "stop " + name

    # Send the command to execute
    out, err = DETAILS["server"].sendline(cmd)

    # "scrape" the output and return the right fields as a dict
    return parse(out)


def service_restart(name):
    """
    Restart a "service" on the ssh server

    .. versionadded:: 2015.8.2
    """
    cmd = "restart " + name

    # Send the command to execute
    out, err = DETAILS["server"].sendline(cmd)

    # "scrape" the output and return the right fields as a dict
    return parse(out)



"""
    Unit tests for the salt.modules.nacl module
"""

import salt.modules.nacl
from tests.support.mock import patch


def test_fips_mode():
    """
    Nacl module does not load when fips_mode is True
    """
    opts = {"fips_mode": True}
    with patch("salt.modules.nacl.__opts__", opts, create=True):
        ret = salt.modules.nacl.__virtual__()
        assert ret == (False, "nacl module not available in FIPS mode")



import os
import urllib.parse

import pytest
import salt.utils.json
import salt.utils.stringutils
import tests.support.cherrypy_testclasses as cptc


class TestAuth(cptc.BaseRestCherryPyTest):
    def test_get_root_noauth(self):
        """
        GET requests to the root URL should not require auth
        """
        request, response = self.request("/")
        self.assertEqual(response.status, "200 OK")

    def test_post_root_auth(self):
        """
        POST requests to the root URL redirect to login
        """
        request, response = self.request("/", method="POST", data={})
        self.assertEqual(response.status, "401 Unauthorized")

    def test_login_noauth(self):
        """
        GET requests to the login URL should not require auth
        """
        request, response = self.request("/login")
        self.assertEqual(response.status, "200 OK")

    def test_webhook_auth(self):
        """
        Requests to the webhook URL require auth by default
        """
        request, response = self.request("/hook", method="POST", data={})
        self.assertEqual(response.status, "401 Unauthorized")


class TestLogin(cptc.BaseRestCherryPyTest):
    auth_creds = (("username", "saltdev"), ("password", "saltdev"), ("eauth", "auto"))

    def test_good_login(self):
        """
        Test logging in
        """
        body = urllib.parse.urlencode(self.auth_creds)
        request, response = self.request(
            "/login",
            method="POST",
            body=body,
            headers={"content-type": "application/x-www-form-urlencoded"},
        )
        self.assertEqual(response.status, "200 OK")
        return response

    def test_bad_login(self):
        """
        Test logging in
        """
        body = urllib.parse.urlencode({"totally": "invalid_creds"})
        request, response = self.request(
            "/login",
            method="POST",
            body=body,
            headers={"content-type": "application/x-www-form-urlencoded"},
        )
        self.assertEqual(response.status, "401 Unauthorized")

    def test_logout(self):
        ret = self.test_good_login()
        token = ret.headers["X-Auth-Token"]

        body = urllib.parse.urlencode({})
        request, response = self.request(
            "/logout",
            method="POST",
            body=body,
            headers={
                "content-type": "application/x-www-form-urlencoded",
                "X-Auth-Token": token,
            },
        )
        self.assertEqual(response.status, "200 OK")


class TestRun(cptc.BaseRestCherryPyTest):
    auth_creds = (
        ("username", "saltdev_auto"),
        ("password", "saltdev"),
        ("eauth", "auto"),
    )

    low = (
        ("client", "local"),
        ("tgt", "*"),
        ("fun", "test.ping"),
    )

    @pytest.mark.slow_test
    def test_run_good_login(self):
        """
        Test the run URL with good auth credentials
        """
        cmd = dict(self.low, **dict(self.auth_creds))
        body = urllib.parse.urlencode(cmd)

        request, response = self.request(
            "/run",
            method="POST",
            body=body,
            headers={"content-type": "application/x-www-form-urlencoded"},
        )
        self.assertEqual(response.status, "200 OK")

    def test_run_bad_login(self):
        """
        Test the run URL with bad auth credentials
        """
        cmd = dict(self.low, **{"totally": "invalid_creds"})
        body = urllib.parse.urlencode(cmd)

        request, response = self.request(
            "/run",
            method="POST",
            body=body,
            headers={"content-type": "application/x-www-form-urlencoded"},
        )
        self.assertEqual(response.status, "401 Unauthorized")

    def test_run_empty_token(self):
        """
        Test the run URL with empty token
        """
        cmd = dict(self.low, **{"token": ""})
        body = urllib.parse.urlencode(cmd)

        request, response = self.request(
            "/run",
            method="POST",
            body=body,
            headers={"content-type": "application/x-www-form-urlencoded"},
        )
        assert response.status == "401 Unauthorized"

    def test_run_empty_token_upercase(self):
        """
        Test the run URL with empty token with upercase characters
        """
        cmd = dict(self.low, **{"ToKen": ""})
        body = urllib.parse.urlencode(cmd)

        request, response = self.request(
            "/run",
            method="POST",
            body=body,
            headers={"content-type": "application/x-www-form-urlencoded"},
        )
        assert response.status == "401 Unauthorized"

    def test_run_wrong_token(self):
        """
        Test the run URL with incorrect token
        """
        cmd = dict(self.low, **{"token": "bad"})
        body = urllib.parse.urlencode(cmd)

        request, response = self.request(
            "/run",
            method="POST",
            body=body,
            headers={"content-type": "application/x-www-form-urlencoded"},
        )
        assert response.status == "401 Unauthorized"

    def test_run_pathname_token(self):
        """
        Test the run URL with path that exists in token
        """
        cmd = dict(self.low, **{"token": os.path.join("etc", "passwd")})
        body = urllib.parse.urlencode(cmd)

        request, response = self.request(
            "/run",
            method="POST",
            body=body,
            headers={"content-type": "application/x-www-form-urlencoded"},
        )
        assert response.status == "401 Unauthorized"

    def test_run_pathname_not_exists_token(self):
        """
        Test the run URL with path that does not exist in token
        """
        cmd = dict(self.low, **{"token": os.path.join("tmp", "doesnotexist")})
        body = urllib.parse.urlencode(cmd)

        request, response = self.request(
            "/run",
            method="POST",
            body=body,
            headers={"content-type": "application/x-www-form-urlencoded"},
        )
        assert response.status == "401 Unauthorized"

    @pytest.mark.slow_test
    def test_run_extra_parameters(self):
        """
        Test the run URL with good auth credentials
        """
        cmd = dict(self.low, **dict(self.auth_creds))
        cmd["id_"] = "someminionname"
        body = urllib.parse.urlencode(cmd)

        request, response = self.request(
            "/run",
            method="POST",
            body=body,
            headers={"content-type": "application/x-www-form-urlencoded"},
        )
        self.assertEqual(response.status, "200 OK")


class TestWebhookDisableAuth(cptc.BaseRestCherryPyTest):
    def __get_opts__(self):
        return {
            "rest_cherrypy": {
                "port": 8000,
                "debug": True,
                "webhook_disable_auth": True,
            },
        }

    def test_webhook_noauth(self):
        """
        Auth can be disabled for requests to the webhook URL
        """
        body = urllib.parse.urlencode({"foo": "Foo!"})
        request, response = self.request(
            "/hook",
            method="POST",
            body=body,
            headers={"content-type": "application/x-www-form-urlencoded"},
        )
        self.assertEqual(response.status, "200 OK")


class TestArgKwarg(cptc.BaseRestCherryPyTest):
    auth_creds = (("username", "saltdev"), ("password", "saltdev"), ("eauth", "auto"))

    low = (
        ("client", "runner"),
        ("fun", "test.arg"),
        # use singular form for arg and kwarg
        ("arg", [1234]),
        ("kwarg", {"ext_source": "redis"}),
    )

    def _token(self):
        """
        Return the token
        """
        body = urllib.parse.urlencode(self.auth_creds)
        request, response = self.request(
            "/login",
            method="POST",
            body=body,
            headers={"content-type": "application/x-www-form-urlencoded"},
        )
        return response.headers["X-Auth-Token"]

    @pytest.mark.slow_test
    def test_accepts_arg_kwarg_keys(self):
        """
        Ensure that (singular) arg and kwarg keys (for passing parameters)
        are supported by runners.
        """
        cmd = dict(self.low)
        body = salt.utils.json.dumps(cmd)

        request, response = self.request(
            "/",
            method="POST",
            body=body,
            headers={
                "content-type": "application/json",
                "X-Auth-Token": self._token(),
                "Accept": "application/json",
            },
        )
        resp = salt.utils.json.loads(salt.utils.stringutils.to_str(response.body[0]))
        self.assertEqual(resp["return"][0]["args"], [1234])
        self.assertEqual(resp["return"][0]["kwargs"], {"ext_source": "redis"})


class TestJobs(cptc.BaseRestCherryPyTest):
    auth_creds = (
        ("username", "saltdev_auto"),
        ("password", "saltdev"),
        ("eauth", "auto"),
    )

    low = (
        ("client", "local"),
        ("tgt", "*"),
        ("fun", "test.ping"),
    )

    def _token(self):
        """
        Return the token
        """
        body = urllib.parse.urlencode(self.auth_creds)
        request, response = self.request(
            "/login",
            method="POST",
            body=body,
            headers={"content-type": "application/x-www-form-urlencoded"},
        )
        return response.headers["X-Auth-Token"]

    def _add_job(self):
        """
        Helper function to add a job to the job cache
        """
        cmd = dict(self.low, **dict(self.auth_creds))
        body = urllib.parse.urlencode(cmd)

        request, response = self.request(
            "/run",
            method="POST",
            body=body,
            headers={"content-type": "application/x-www-form-urlencoded"},
        )
        self.assertEqual(response.status, "200 OK")

    @pytest.mark.flaky(max_runs=4)
    @pytest.mark.slow_test
    def test_all_jobs(self):
        """
        test query to /jobs returns job data
        """
        self._add_job()

        request, response = self.request(
            "/jobs",
            method="GET",
            headers={"Accept": "application/json", "X-Auth-Token": self._token()},
        )

        resp = salt.utils.json.loads(salt.utils.stringutils.to_str(response.body[0]))
        self.assertIn("test.ping", str(resp["return"]))
        self.assertEqual(response.status, "200 OK")



"""
Module to manage filesystem snapshots with snapper

.. versionadded:: 2016.11.0

:codeauthor:    Duncan Mac-Vicar P. <dmacvicar@suse.de>
:codeauthor:    Pablo Suárez Hernández <psuarezhernandez@suse.de>

:depends:       ``dbus`` Python module.
:depends:       ``snapper`` http://snapper.io, available in most distros
:maturity:      new
:platform:      Linux
"""
import difflib
import logging
import os
import subprocess
import time

import salt.utils.files
from salt.exceptions import CommandExecutionError

try:
    from pwd import getpwuid

    HAS_PWD = True
except ImportError:
    HAS_PWD = False


try:
    import dbus  # pylint: disable=wrong-import-order

    HAS_DBUS = True
except ImportError:
    HAS_DBUS = False


DBUS_STATUS_MAP = {
    1: "created",
    2: "deleted",
    4: "type changed",
    8: "modified",
    16: "permission changed",
    32: "owner changed",
    64: "group changed",
    128: "extended attributes changed",
    256: "ACL info changed",
}

SNAPPER_DBUS_OBJECT = "org.opensuse.Snapper"
SNAPPER_DBUS_PATH = "/org/opensuse/Snapper"
SNAPPER_DBUS_INTERFACE = "org.opensuse.Snapper"

log = logging.getLogger(__name__)

# pylint: disable=invalid-name
bus = None
system_bus_error = None
snapper = None
snapper_error = None

if HAS_DBUS:
    try:
        bus = dbus.SystemBus()
    except dbus.DBusException as exc:
        system_bus_error = exc
    else:
        if SNAPPER_DBUS_OBJECT in bus.list_activatable_names():
            try:
                snapper = dbus.Interface(
                    bus.get_object(SNAPPER_DBUS_OBJECT, SNAPPER_DBUS_PATH),
                    dbus_interface=SNAPPER_DBUS_INTERFACE,
                )
            except (dbus.DBusException, ValueError) as exc:
                snapper_error = exc
        else:
            snapper_error = "snapper is missing"
# pylint: enable=invalid-name


def __virtual__():
    error_msg = "The snapper module cannot be loaded: {0}"
    if not HAS_DBUS:
        return False, error_msg.format("missing python dbus module")
    elif not snapper:
        return False, error_msg.format(snapper_error)
    elif not bus:
        return False, error_msg.format(system_bus_error)
    elif not HAS_PWD:
        return False, error_msg.format("pwd module not available")

    return "snapper"


def _snapshot_to_data(snapshot):
    """
    Returns snapshot data from a D-Bus response.

    A snapshot D-Bus response is a dbus.Struct containing the
    information related to a snapshot:

    [id, type, pre_snapshot, timestamp, user, description,
     cleanup_algorithm, userdata]

    id: dbus.UInt32
    type: dbus.UInt16
    pre_snapshot: dbus.UInt32
    timestamp: dbus.Int64
    user: dbus.UInt32
    description: dbus.String
    cleaup_algorithm: dbus.String
    userdata: dbus.Dictionary
    """
    data = {}

    data["id"] = snapshot[0]
    data["type"] = ["single", "pre", "post"][snapshot[1]]
    if data["type"] == "post":
        data["pre"] = snapshot[2]

    if snapshot[3] != -1:
        data["timestamp"] = snapshot[3]
    else:
        data["timestamp"] = int(time.time())

    data["user"] = getpwuid(snapshot[4])[0]
    data["description"] = snapshot[5]
    data["cleanup"] = snapshot[6]

    data["userdata"] = {}
    for key, value in snapshot[7].items():
        data["userdata"][key] = value

    return data


def _dbus_exception_to_reason(exc, args):
    """
    Returns a error message from a snapper DBusException
    """
    error = exc.get_dbus_name()
    if error == "error.unknown_config":
        return "Unknown configuration '{}'".format(args["config"])
    elif error == "error.illegal_snapshot":
        return "Invalid snapshot"
    else:
        return exc.get_dbus_name()


def list_snapshots(config="root"):
    """
    List available snapshots

    CLI Example:

    .. code-block:: bash

        salt '*' snapper.list_snapshots config=myconfig
    """
    try:
        snapshots = snapper.ListSnapshots(config)
        return [_snapshot_to_data(s) for s in snapshots]
    except dbus.DBusException as exc:
        raise CommandExecutionError(
            "Error encountered while listing snapshots: {}".format(
                _dbus_exception_to_reason(exc, locals())
            )
        )


def get_snapshot(number=0, config="root"):
    """
    Get detailed information about a given snapshot

    CLI Example:

    .. code-block:: bash

        salt '*' snapper.get_snapshot 1
    """
    try:
        snapshot = snapper.GetSnapshot(config, int(number))
        return _snapshot_to_data(snapshot)
    except dbus.DBusException as exc:
        raise CommandExecutionError(
            "Error encountered while retrieving snapshot: {}".format(
                _dbus_exception_to_reason(exc, locals())
            )
        )


def list_configs():
    """
    List all available configs

    CLI Example:

    .. code-block:: bash

        salt '*' snapper.list_configs
    """
    try:
        configs = snapper.ListConfigs()
        return {config[0]: config[2] for config in configs}
    except dbus.DBusException as exc:
        raise CommandExecutionError(
            "Error encountered while listing configurations: {}".format(
                _dbus_exception_to_reason(exc, locals())
            )
        )


def _config_filter(value):
    if isinstance(value, bool):
        return "yes" if value else "no"
    return value


def set_config(name="root", **kwargs):
    """
    Set configuration values

    CLI Example:

    .. code-block:: bash

        salt '*' snapper.set_config SYNC_ACL=True

    Keys are case insensitive as they will be always uppercased to snapper
    convention. The above example is equivalent to:

    .. code-block:: bash

        salt '*' snapper.set_config sync_acl=True
    """
    try:
        data = {
            k.upper(): _config_filter(v)
            for k, v in kwargs.items()
            if not k.startswith("__")
        }
        snapper.SetConfig(name, data)
    except dbus.DBusException as exc:
        raise CommandExecutionError(
            "Error encountered while setting configuration {}: {}".format(
                name, _dbus_exception_to_reason(exc, locals())
            )
        )
    return True


def _get_last_snapshot(config="root"):
    """
    Returns the last existing created snapshot
    """
    snapshot_list = sorted(list_snapshots(config), key=lambda x: x["id"])
    return snapshot_list[-1]


def status_to_string(dbus_status):
    """
    Converts a numeric dbus snapper status into a string

    CLI Example:

    .. code-block:: bash

        salt '*' snapper.status_to_string <dbus_status>
    """
    status_tuple = (
        dbus_status & 0b000000001,
        dbus_status & 0b000000010,
        dbus_status & 0b000000100,
        dbus_status & 0b000001000,
        dbus_status & 0b000010000,
        dbus_status & 0b000100000,
        dbus_status & 0b001000000,
        dbus_status & 0b010000000,
        dbus_status & 0b100000000,
    )

    return [DBUS_STATUS_MAP[status] for status in status_tuple if status]


def get_config(name="root"):
    """
    Retrieves all values from a given configuration

    CLI Example:

    .. code-block:: bash

      salt '*' snapper.get_config
    """
    try:
        config = snapper.GetConfig(name)
        return config
    except dbus.DBusException as exc:
        raise CommandExecutionError(
            "Error encountered while retrieving configuration: {}".format(
                _dbus_exception_to_reason(exc, locals())
            )
        )


def create_config(
    name=None, subvolume=None, fstype=None, template=None, extra_opts=None
):
    """
    Creates a new Snapper configuration

    name
        Name of the new Snapper configuration.
    subvolume
        Path to the related subvolume.
    fstype
        Filesystem type of the subvolume.
    template
        Configuration template to use. (Default: default)
    extra_opts
        Extra Snapper configuration opts dictionary. It will override the values provided
        by the given template (if any).

    CLI Example:

    .. code-block:: bash

      salt '*' snapper.create_config name=myconfig subvolume=/foo/bar/ fstype=btrfs
      salt '*' snapper.create_config name=myconfig subvolume=/foo/bar/ fstype=btrfs template="default"
      salt '*' snapper.create_config name=myconfig subvolume=/foo/bar/ fstype=btrfs extra_opts='{"NUMBER_CLEANUP": False}'
    """

    def raise_arg_error(argname):
        raise CommandExecutionError(
            'You must provide a "{}" for the new configuration'.format(argname)
        )

    if not name:
        raise_arg_error("name")
    if not subvolume:
        raise_arg_error("subvolume")
    if not fstype:
        raise_arg_error("fstype")
    if not template:
        template = ""

    try:
        snapper.CreateConfig(name, subvolume, fstype, template)
        if extra_opts:
            set_config(name, **extra_opts)
        return get_config(name)
    except dbus.DBusException as exc:
        raise CommandExecutionError(
            "Error encountered while creating the new configuration: {}".format(
                _dbus_exception_to_reason(exc, locals())
            )
        )


def create_snapshot(
    config="root",
    snapshot_type="single",
    pre_number=None,
    description=None,
    cleanup_algorithm="number",
    userdata=None,
    **kwargs
):
    """
    Creates an snapshot

    config
        Configuration name.
    snapshot_type
        Specifies the type of the new snapshot. Possible values are
        single, pre and post.
    pre_number
        For post snapshots the number of the pre snapshot must be
        provided.
    description
        Description for the snapshot. If not given, the salt job will be used.
    cleanup_algorithm
        Set the cleanup algorithm for the snapshot.

    number
        Deletes old snapshots when a certain number of snapshots
        is reached.
    timeline
        Deletes old snapshots but keeps a number of hourly,
        daily, weekly, monthly and yearly snapshots.
    empty-pre-post
        Deletes pre/post snapshot pairs with empty diffs.
    userdata
        Set userdata for the snapshot (key-value pairs).

    Returns the number of the created snapshot.

    CLI Example:

    .. code-block:: bash

        salt '*' snapper.create_snapshot
    """
    if not userdata:
        userdata = {}

    jid = kwargs.get("__pub_jid")
    if description is None and jid is not None:
        description = "salt job {}".format(jid)

    if jid is not None:
        userdata["salt_jid"] = jid

    new_nr = None
    try:
        if snapshot_type == "single":
            new_nr = snapper.CreateSingleSnapshot(
                config, description, cleanup_algorithm, userdata
            )
        elif snapshot_type == "pre":
            new_nr = snapper.CreatePreSnapshot(
                config, description, cleanup_algorithm, userdata
            )
        elif snapshot_type == "post":
            if pre_number is None:
                raise CommandExecutionError(
                    "pre snapshot number 'pre_number' needs to be"
                    "specified for snapshots of the 'post' type"
                )
            new_nr = snapper.CreatePostSnapshot(
                config, pre_number, description, cleanup_algorithm, userdata
            )
        else:
            raise CommandExecutionError(
                "Invalid snapshot type '{}'".format(snapshot_type)
            )
    except dbus.DBusException as exc:
        raise CommandExecutionError(
            "Error encountered while listing changed files: {}".format(
                _dbus_exception_to_reason(exc, locals())
            )
        )
    return new_nr


def delete_snapshot(snapshots_ids=None, config="root"):
    """
    Deletes an snapshot

    config
        Configuration name. (Default: root)

    snapshots_ids
        List of the snapshots IDs to be deleted.

    CLI Example:

    .. code-block:: bash

        salt '*' snapper.delete_snapshot 54
        salt '*' snapper.delete_snapshot config=root 54
        salt '*' snapper.delete_snapshot config=root snapshots_ids=[54,55,56]
    """
    if not snapshots_ids:
        raise CommandExecutionError("Error: No snapshot ID has been provided")
    try:
        current_snapshots_ids = [x["id"] for x in list_snapshots(config)]
        if not isinstance(snapshots_ids, list):
            snapshots_ids = [snapshots_ids]
        if not set(snapshots_ids).issubset(set(current_snapshots_ids)):
            raise CommandExecutionError(
                "Error: Snapshots '{}' not found".format(
                    ", ".join(
                        [
                            str(x)
                            for x in set(snapshots_ids).difference(
                                set(current_snapshots_ids)
                            )
                        ]
                    )
                )
            )
        snapper.DeleteSnapshots(config, snapshots_ids)
        return {config: {"ids": snapshots_ids, "status": "deleted"}}
    except dbus.DBusException as exc:
        raise CommandExecutionError(_dbus_exception_to_reason(exc, locals()))


def modify_snapshot(
    snapshot_id=None, description=None, userdata=None, cleanup=None, config="root"
):
    """
    Modify attributes of an existing snapshot.

    config
        Configuration name. (Default: root)

    snapshot_id
        ID of the snapshot to be modified.

    cleanup
        Change the cleanup method of the snapshot. (str)

    description
        Change the description of the snapshot. (str)

    userdata
        Change the userdata dictionary of the snapshot. (dict)

    CLI Example:

    .. code-block:: bash

        salt '*' snapper.modify_snapshot 54 description="my snapshot description"
        salt '*' snapper.modify_snapshot 54 description="my snapshot description"
        salt '*' snapper.modify_snapshot 54 userdata='{"foo": "bar"}'
        salt '*' snapper.modify_snapshot snapshot_id=54 cleanup="number"
    """
    if not snapshot_id:
        raise CommandExecutionError("Error: No snapshot ID has been provided")

    snapshot = get_snapshot(config=config, number=snapshot_id)
    try:
        # Updating only the explicitly provided attributes by the user
        updated_opts = {
            "description": description
            if description is not None
            else snapshot["description"],
            "cleanup": cleanup if cleanup is not None else snapshot["cleanup"],
            "userdata": userdata if userdata is not None else snapshot["userdata"],
        }
        snapper.SetSnapshot(
            config,
            snapshot_id,
            updated_opts["description"],
            updated_opts["cleanup"],
            updated_opts["userdata"],
        )
        return get_snapshot(config=config, number=snapshot_id)
    except dbus.DBusException as exc:
        raise CommandExecutionError(_dbus_exception_to_reason(exc, locals()))


def _get_num_interval(config, num_pre, num_post):
    """
    Returns numerical interval based on optionals num_pre, num_post values
    """
    post = int(num_post) if num_post else 0
    pre = int(num_pre) if num_pre is not None else _get_last_snapshot(config)["id"]
    return pre, post


def _is_text_file(filename):
    """
    Checks if a file is a text file
    """
    type_of_file = subprocess.run(
        ["file", "-bi", filename],
        check=False,
        stdout=subprocess.PIPE,
        universal_newlines=True,
    ).stdout
    return type_of_file.startswith("text")


def run(function, *args, **kwargs):
    """
    Runs a function from an execution module creating pre and post snapshots
    and associating the salt job id with those snapshots for easy undo and
    cleanup.

    function
        Salt function to call.

    config
        Configuration name. (default: "root")

    description
        A description for the snapshots. (default: None)

    userdata
        Data to include in the snapshot metadata. (default: None)

    cleanup_algorithm
        Snapper cleanup algorithm. (default: "number")

    `*args`
        args for the function to call. (default: None)

    `**kwargs`
        kwargs for the function to call (default: None)

    This  would run append text to /etc/motd using the file.append
    module, and will create two snapshots, pre and post with the associated
    metadata. The jid will be available as salt_jid in the userdata of the
    snapshot.

    You can immediately see the changes

    CLI Example:

    .. code-block:: bash

        salt '*' snapper.run file.append args='["/etc/motd", "some text"]'
    """
    config = kwargs.pop("config", "root")
    description = kwargs.pop("description", "snapper.run[{}]".format(function))
    cleanup_algorithm = kwargs.pop("cleanup_algorithm", "number")
    userdata = kwargs.pop("userdata", {})

    func_kwargs = {k: v for k, v in kwargs.items() if not k.startswith("__")}
    kwargs = {k: v for k, v in kwargs.items() if k.startswith("__")}

    pre_nr = __salt__["snapper.create_snapshot"](
        config=config,
        snapshot_type="pre",
        description=description,
        cleanup_algorithm=cleanup_algorithm,
        userdata=userdata,
        **kwargs
    )

    if function not in __salt__:
        raise CommandExecutionError('function "{}" does not exist'.format(function))

    try:
        ret = __salt__[function](*args, **func_kwargs)
    except CommandExecutionError as exc:
        ret = "\n".join([str(exc), __salt__[function].__doc__])

    __salt__["snapper.create_snapshot"](
        config=config,
        snapshot_type="post",
        pre_number=pre_nr,
        description=description,
        cleanup_algorithm=cleanup_algorithm,
        userdata=userdata,
        **kwargs
    )
    return ret


def status(config="root", num_pre=None, num_post=None):
    """
    Returns a comparison between two snapshots

    config
        Configuration name.

    num_pre
        first snapshot ID to compare. Default is last snapshot

    num_post
        last snapshot ID to compare. Default is 0 (current state)

    CLI Example:

    .. code-block:: bash

        salt '*' snapper.status
        salt '*' snapper.status num_pre=19 num_post=20
    """
    try:
        pre, post = _get_num_interval(config, num_pre, num_post)
        snapper.CreateComparison(config, int(pre), int(post))
        files = snapper.GetFiles(config, int(pre), int(post))
        status_ret = {}
        subvolume = list_configs()[config]["SUBVOLUME"]
        for file in files:
            # In case of SUBVOLUME is included in filepath we remove it
            # to prevent from filepath starting with double '/'
            _filepath = (
                file[0][len(subvolume) :] if file[0].startswith(subvolume) else file[0]
            )
            status_ret[os.path.normpath(subvolume + _filepath)] = {
                "status": status_to_string(file[1])
            }
        return status_ret
    except dbus.DBusException as exc:
        raise CommandExecutionError(
            "Error encountered while listing changed files: {}".format(
                _dbus_exception_to_reason(exc, locals())
            )
        )


def changed_files(config="root", num_pre=None, num_post=None):
    """
    Returns the files changed between two snapshots

    config
        Configuration name.

    num_pre
        first snapshot ID to compare. Default is last snapshot

    num_post
        last snapshot ID to compare. Default is 0 (current state)

    CLI Example:

    .. code-block:: bash

        salt '*' snapper.changed_files
        salt '*' snapper.changed_files num_pre=19 num_post=20
    """
    return status(config, num_pre, num_post).keys()


def undo(config="root", files=None, num_pre=None, num_post=None):
    """
    Undo all file changes that happened between num_pre and num_post, leaving
    the files into the state of num_pre.

    .. warning::
        If one of the files has changes after num_post, they will be overwritten
        The snapshots are used to determine the file list, but the current
        version of the files will be overwritten by the versions in num_pre.

        You to undo changes between num_pre and the current version of the
        files use num_post=0.

    CLI Example:

    .. code-block:: bash

        salt '*' snapper.undo
    """
    pre, post = _get_num_interval(config, num_pre, num_post)

    changes = status(config, pre, post)
    changed = set(changes.keys())
    requested = set(files or changed)

    if not requested.issubset(changed):
        raise CommandExecutionError(
            "Given file list contains files that are not present"
            "in the changed filelist: {}".format(changed - requested)
        )

    cmdret = __salt__["cmd.run"](
        "snapper -c {} undochange {}..{} {}".format(
            config, pre, post, " ".join(requested)
        )
    )

    try:
        components = cmdret.split(" ")
        ret = {}
        for comp in components:
            key, val = comp.split(":")
            ret[key] = val
        return ret
    except ValueError as exc:
        raise CommandExecutionError(
            "Error while processing Snapper response: {}".format(cmdret)
        )


def _get_jid_snapshots(jid, config="root"):
    """
    Returns pre/post snapshots made by a given Salt jid

    Looks for 'salt_jid' entries into snapshots userdata which are created
    when 'snapper.run' is executed.
    """
    jid_snapshots = [
        x for x in list_snapshots(config) if x["userdata"].get("salt_jid") == jid
    ]
    pre_snapshot = [x for x in jid_snapshots if x["type"] == "pre"]
    post_snapshot = [x for x in jid_snapshots if x["type"] == "post"]

    if not pre_snapshot or not post_snapshot:
        raise CommandExecutionError("Jid '{}' snapshots not found".format(jid))

    return (pre_snapshot[0]["id"], post_snapshot[0]["id"])


def undo_jid(jid, config="root"):
    """
    Undo the changes applied by a salt job

    jid
        The job id to lookup

    config
        Configuration name.

    CLI Example:

    .. code-block:: bash

        salt '*' snapper.undo_jid jid=20160607130930720112
    """
    pre_snapshot, post_snapshot = _get_jid_snapshots(jid, config=config)
    return undo(config, num_pre=pre_snapshot, num_post=post_snapshot)


def diff(config="root", filename=None, num_pre=None, num_post=None):
    """
    Returns the differences between two snapshots

    config
        Configuration name.

    filename
        if not provided the showing differences between snapshots for
        all "text" files

    num_pre
        first snapshot ID to compare. Default is last snapshot

    num_post
        last snapshot ID to compare. Default is 0 (current state)

    CLI Example:

    .. code-block:: bash

        salt '*' snapper.diff
        salt '*' snapper.diff filename=/var/log/snapper.log num_pre=19 num_post=20
    """
    try:
        pre, post = _get_num_interval(config, num_pre, num_post)

        files = changed_files(config, pre, post)
        if filename:
            files = [filename] if filename in files else []

        subvolume = list_configs()[config]["SUBVOLUME"]
        pre_mount = snapper.MountSnapshot(config, pre, False) if pre else subvolume
        post_mount = snapper.MountSnapshot(config, post, False) if post else subvolume

        files_diff = dict()
        for filepath in [filepath for filepath in files if not os.path.isdir(filepath)]:

            _filepath = filepath
            if filepath.startswith(subvolume):
                _filepath = filepath[len(subvolume) :]

            # Just in case, removing possible double '/' from the final file paths
            pre_file = os.path.normpath(pre_mount + "/" + _filepath).replace("//", "/")
            post_file = os.path.normpath(post_mount + "/" + _filepath).replace(
                "//", "/"
            )

            if os.path.isfile(pre_file):
                pre_file_exists = True
                with salt.utils.files.fopen(pre_file) as rfh:
                    pre_file_content = [
                        salt.utils.stringutils.to_unicode(_l) for _l in rfh.readlines()
                    ]
            else:
                pre_file_content = []
                pre_file_exists = False

            if os.path.isfile(post_file):
                post_file_exists = True
                with salt.utils.files.fopen(post_file) as rfh:
                    post_file_content = [
                        salt.utils.stringutils.to_unicode(_l) for _l in rfh.readlines()
                    ]
            else:
                post_file_content = []
                post_file_exists = False

            if _is_text_file(pre_file) or _is_text_file(post_file):
                files_diff[filepath] = {
                    "comment": "text file changed",
                    "diff": "".join(
                        difflib.unified_diff(
                            pre_file_content,
                            post_file_content,
                            fromfile=pre_file,
                            tofile=post_file,
                        )
                    ),
                }

                if pre_file_exists and not post_file_exists:
                    files_diff[filepath]["comment"] = "text file deleted"
                if not pre_file_exists and post_file_exists:
                    files_diff[filepath]["comment"] = "text file created"

            elif not _is_text_file(pre_file) and not _is_text_file(post_file):
                # This is a binary file
                files_diff[filepath] = {"comment": "binary file changed"}
                if pre_file_exists:
                    files_diff[filepath]["old_sha256_digest"] = __salt__[
                        "hashutil.sha256_digest"
                    ]("".join(pre_file_content))
                if post_file_exists:
                    files_diff[filepath]["new_sha256_digest"] = __salt__[
                        "hashutil.sha256_digest"
                    ]("".join(post_file_content))
                if post_file_exists and not pre_file_exists:
                    files_diff[filepath]["comment"] = "binary file created"
                if pre_file_exists and not post_file_exists:
                    files_diff[filepath]["comment"] = "binary file deleted"

        if pre:
            snapper.UmountSnapshot(config, pre, False)
        if post:
            snapper.UmountSnapshot(config, post, False)
        return files_diff
    except dbus.DBusException as exc:
        raise CommandExecutionError(
            "Error encountered while showing differences between snapshots: {}".format(
                _dbus_exception_to_reason(exc, locals())
            )
        )


def diff_jid(jid, config="root"):
    """
    Returns the changes applied by a `jid`

    jid
        The job id to lookup

    config
        Configuration name.

    CLI Example:

    .. code-block:: bash

        salt '*' snapper.diff_jid jid=20160607130930720112
    """
    pre_snapshot, post_snapshot = _get_jid_snapshots(jid, config=config)
    return diff(config, num_pre=pre_snapshot, num_post=post_snapshot)


def create_baseline(tag="baseline", config="root"):
    """
    Creates a snapshot marked as baseline

    tag
        Tag name for the baseline

    config
        Configuration name.

    CLI Example:

    .. code-block:: bash

        salt '*' snapper.create_baseline
        salt '*' snapper.create_baseline my_custom_baseline
    """
    return __salt__["snapper.create_snapshot"](
        config=config,
        snapshot_type="single",
        description="baseline snapshot",
        cleanup_algorithm="number",
        userdata={"baseline_tag": tag},
    )



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""


import salt.modules.useradd as useradd
from salt.exceptions import CommandExecutionError
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase, skipIf

try:
    import pwd

    HAS_PWD = True
except ImportError:
    HAS_PWD = False


class UserAddTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.useradd
    """

    def setup_loader_modules(self):
        return {useradd: {}}

    @classmethod
    def setUpClass(cls):
        cls.mock_pwall = {
            "gid": 0,
            "groups": ["root"],
            "home": "/root",
            "name": "root",
            "passwd": "x",
            "shell": "/bin/bash",
            "uid": 0,
            "fullname": "root",
            "roomnumber": "",
            "workphone": "",
            "homephone": "",
            "other": "",
        }

    @classmethod
    def tearDownClass(cls):
        del cls.mock_pwall

    # 'add' function tests: 1

    def test_add(self):
        """
        Test for adding a user
        """
        with patch.dict(useradd.__grains__, {"kernel": "OpenBSD"}):
            mock = MagicMock(return_value={"retcode": 0})
            with patch.dict(useradd.__salt__, {"cmd.run_all": mock}):
                self.assertTrue(useradd.add("Salt"))

            mock = MagicMock(return_value={"retcode": 1})
            with patch.dict(useradd.__salt__, {"cmd.run_all": mock}):
                self.assertFalse(useradd.add("Salt"))

    # 'getent' function tests: 2

    @skipIf(HAS_PWD is False, "The pwd module is not available")
    def test_getent(self):
        """
        Test if user.getent already have a value
        """
        with patch("salt.modules.useradd.__context__", MagicMock(return_value="Salt")):
            self.assertTrue(useradd.getent())

    @skipIf(HAS_PWD is False, "The pwd module is not available")
    def test_getent_user(self):
        """
        Tests the return information on all users
        """
        with patch("pwd.getpwall", MagicMock(return_value=[""])):
            ret = [
                {
                    "gid": 0,
                    "groups": ["root"],
                    "home": "/root",
                    "name": "root",
                    "passwd": "x",
                    "shell": "/bin/bash",
                    "uid": 0,
                    "fullname": "root",
                    "roomnumber": "",
                    "workphone": "",
                    "homephone": "",
                    "other": "",
                }
            ]
            with patch(
                "salt.modules.useradd._format_info",
                MagicMock(return_value=self.mock_pwall),
            ):
                self.assertEqual(useradd.getent(), ret)

    # 'chuid' function tests: 1

    def test_chuid(self):
        """
        Test if the uid of a user change
        """
        mock = MagicMock(return_value={"uid": 11})
        with patch.object(useradd, "info", mock):
            self.assertTrue(useradd.chuid("name", 11))

        mock_run = MagicMock(return_value=None)
        with patch.dict(useradd.__salt__, {"cmd.run": mock_run}):
            mock = MagicMock(side_effect=[{"uid": 11}, {"uid": 11}])
            with patch.object(useradd, "info", mock):
                self.assertFalse(useradd.chuid("name", 22))

        with patch.dict(useradd.__salt__, {"cmd.run": mock_run}):
            mock = MagicMock(side_effect=[{"uid": 11}, {"uid": 22}])
            with patch.object(useradd, "info", mock):
                self.assertTrue(useradd.chuid("name", 11))

    # 'chgid' function tests: 1

    def test_chgid(self):
        """
        Test the default group of the user
        """
        mock = MagicMock(return_value={"gid": 11})
        with patch.object(useradd, "info", mock):
            self.assertTrue(useradd.chgid("name", 11))

        mock_run = MagicMock(return_value=None)
        with patch.dict(useradd.__salt__, {"cmd.run": mock_run}):
            mock = MagicMock(side_effect=[{"gid": 22}, {"gid": 22}])
            with patch.object(useradd, "info", mock):
                self.assertFalse(useradd.chgid("name", 11))

        with patch.dict(useradd.__salt__, {"cmd.run": mock_run}):
            mock = MagicMock(side_effect=[{"gid": 11}, {"gid": 22}])
            with patch.object(useradd, "info", mock):
                self.assertTrue(useradd.chgid("name", 11))

    # 'chshell' function tests: 1

    def test_chshell(self):
        """
        Test the default shell of user
        """
        mock = MagicMock(return_value={"shell": "/bin/bash"})
        with patch.object(useradd, "info", mock):
            self.assertTrue(useradd.chshell("name", "/bin/bash"))

        mock_run = MagicMock(return_value=None)
        with patch.dict(useradd.__salt__, {"cmd.run": mock_run}):
            mock = MagicMock(
                side_effect=[{"shell": "/bin/bash"}, {"shell": "/bin/bash"}]
            )
            with patch.object(useradd, "info", mock):
                self.assertFalse(useradd.chshell("name", "/usr/bash"))

        with patch.dict(useradd.__salt__, {"cmd.run": mock_run}):
            mock = MagicMock(
                side_effect=[{"shell": "/bin/bash"}, {"shell": "/usr/bash"}]
            )
            with patch.object(useradd, "info", mock):
                self.assertTrue(useradd.chshell("name", "/bin/bash"))

    # 'chhome' function tests: 1

    def test_chhome(self):
        """
        Test if home directory given is same as previous home directory
        """
        mock = MagicMock(return_value={"home": "/root"})
        with patch.object(useradd, "info", mock):
            self.assertTrue(useradd.chhome("name", "/root"))

        mock = MagicMock(return_value=None)
        with patch.dict(useradd.__salt__, {"cmd.run": mock}):
            mock = MagicMock(side_effect=[{"home": "/root"}, {"home": "/root"}])
            with patch.object(useradd, "info", mock):
                self.assertFalse(useradd.chhome("name", "/user"))

        mock = MagicMock(return_value=None)
        with patch.dict(useradd.__salt__, {"cmd.run": mock}):
            mock = MagicMock(side_effect=[{"home": "/root"}, {"home": "/root"}])
            with patch.object(useradd, "info", mock):
                self.assertTrue(useradd.chhome("name", "/root"))

    # 'chgroups' function tests: 1

    def test_chgroups(self):
        """
        Test if user groups changed
        """
        mock = MagicMock(return_value=["wheel", "root"])
        with patch.object(useradd, "list_groups", mock):
            self.assertTrue(useradd.chgroups("foo", "wheel,root"))

        mock = MagicMock(return_value=["wheel", "root"])
        with patch.object(useradd, "list_groups", mock):
            with patch.dict(useradd.__grains__, {"kernel": "OpenBSD"}):
                mock_runall = MagicMock(return_value={"retcode": False, "stderr": ""})
                with patch.dict(useradd.__salt__, {"cmd.run_all": mock_runall}):
                    self.assertTrue(useradd.chgroups("foo", "wheel,test,root"))

                mock_runall = MagicMock(return_value={"retcode": True, "stderr": ""})
                with patch.dict(useradd.__salt__, {"cmd.run_all": mock_runall}):
                    self.assertFalse(useradd.chgroups("foo", "wheel,test,root"))

    # 'chfullname' function tests: 1

    def test_chfullname(self):
        """
        Test if the user's Full Name is changed
        """
        mock = MagicMock(return_value=False)
        with patch.object(useradd, "_get_gecos", mock):
            self.assertFalse(useradd.chfullname("Salt", "SaltStack"))

        mock = MagicMock(return_value={"fullname": "SaltStack"})
        with patch.object(useradd, "_get_gecos", mock):
            self.assertTrue(useradd.chfullname("Salt", "SaltStack"))

        mock = MagicMock(return_value={"fullname": "SaltStack"})
        with patch.object(useradd, "_get_gecos", mock):
            mock = MagicMock(return_value=None)
            with patch.dict(useradd.__salt__, {"cmd.run": mock}):
                mock = MagicMock(return_value={"fullname": "SaltStack2"})
                with patch.object(useradd, "info", mock):
                    self.assertFalse(useradd.chfullname("Salt", "SaltStack1"))

        mock = MagicMock(return_value={"fullname": "SaltStack2"})
        with patch.object(useradd, "_get_gecos", mock):
            mock = MagicMock(return_value=None)
            with patch.dict(useradd.__salt__, {"cmd.run": mock}):
                mock = MagicMock(return_value={"fullname": "SaltStack2"})
                with patch.object(useradd, "info", mock):
                    self.assertFalse(useradd.chfullname("Salt", "SaltStack1"))

    # 'chroomnumber' function tests: 1

    def test_chroomnumber(self):
        """
        Test if the user's Room Number is changed
        """
        mock = MagicMock(return_value=False)
        with patch.object(useradd, "_get_gecos", mock):
            self.assertFalse(useradd.chroomnumber("salt", 1))

        mock = MagicMock(return_value={"roomnumber": "1"})
        with patch.object(useradd, "_get_gecos", mock):
            self.assertTrue(useradd.chroomnumber("salt", 1))

        mock = MagicMock(return_value={"roomnumber": "2"})
        with patch.object(useradd, "_get_gecos", mock):
            mock = MagicMock(return_value=None)
            with patch.dict(useradd.__salt__, {"cmd.run": mock}):
                mock = MagicMock(return_value={"roomnumber": "3"})
                with patch.object(useradd, "info", mock):
                    self.assertFalse(useradd.chroomnumber("salt", 1))

        mock = MagicMock(return_value={"roomnumber": "3"})
        with patch.object(useradd, "_get_gecos", mock):
            mock = MagicMock(return_value=None)
            with patch.dict(useradd.__salt__, {"cmd.run": mock}):
                mock = MagicMock(return_value={"roomnumber": "3"})
                with patch.object(useradd, "info", mock):
                    self.assertFalse(useradd.chroomnumber("salt", 1))

    # 'chworkphone' function tests: 1

    def test_chworkphone(self):
        """
        Test if the user's Work Phone is changed
        """
        mock = MagicMock(return_value=False)
        with patch.object(useradd, "_get_gecos", mock):
            self.assertFalse(useradd.chworkphone("salt", 1))

        mock = MagicMock(return_value={"workphone": "1"})
        with patch.object(useradd, "_get_gecos", mock):
            self.assertTrue(useradd.chworkphone("salt", 1))

        mock = MagicMock(return_value={"workphone": "2"})
        with patch.object(useradd, "_get_gecos", mock):
            mock = MagicMock(return_value=None)
            with patch.dict(useradd.__salt__, {"cmd.run": mock}):
                mock = MagicMock(return_value={"workphone": "3"})
                with patch.object(useradd, "info", mock):
                    self.assertFalse(useradd.chworkphone("salt", 1))

        mock = MagicMock(return_value={"workphone": "3"})
        with patch.object(useradd, "_get_gecos", mock):
            mock = MagicMock(return_value=None)
            with patch.dict(useradd.__salt__, {"cmd.run": mock}):
                mock = MagicMock(return_value={"workphone": "3"})
                with patch.object(useradd, "info", mock):
                    self.assertFalse(useradd.chworkphone("salt", 1))

    # 'chhomephone' function tests: 1

    def test_chhomephone(self):
        """
        Test if the user's Home Phone is changed
        """
        mock = MagicMock(return_value=False)
        with patch.object(useradd, "_get_gecos", mock):
            self.assertFalse(useradd.chhomephone("salt", 1))

        mock = MagicMock(return_value={"homephone": "1"})
        with patch.object(useradd, "_get_gecos", mock):
            self.assertTrue(useradd.chhomephone("salt", 1))

        mock = MagicMock(return_value={"homephone": "2"})
        with patch.object(useradd, "_get_gecos", mock):
            mock = MagicMock(return_value=None)
            with patch.dict(useradd.__salt__, {"cmd.run": mock}):
                mock = MagicMock(return_value={"homephone": "3"})
                with patch.object(useradd, "info", mock):
                    self.assertFalse(useradd.chhomephone("salt", 1))

        mock = MagicMock(return_value={"homephone": "3"})
        with patch.object(useradd, "_get_gecos", mock):
            mock = MagicMock(return_value=None)
            with patch.dict(useradd.__salt__, {"cmd.run": mock}):
                mock = MagicMock(return_value={"homephone": "3"})
                with patch.object(useradd, "info", mock):
                    self.assertFalse(useradd.chhomephone("salt", 1))

    # 'chother' function tests: 1

    def test_chother(self):
        """
        Test if the user's other GECOS attribute is changed
        """
        mock = MagicMock(return_value=False)
        with patch.object(useradd, "_get_gecos", mock):
            self.assertFalse(useradd.chother("salt", 1))

        mock = MagicMock(return_value={"other": "foobar"})
        with patch.object(useradd, "_get_gecos", mock):
            self.assertTrue(useradd.chother("salt", "foobar"))

        mock = MagicMock(return_value={"other": "foobar2"})
        with patch.object(useradd, "_get_gecos", mock):
            mock = MagicMock(return_value=None)
            with patch.dict(useradd.__salt__, {"cmd.run": mock}):
                mock = MagicMock(return_value={"other": "foobar3"})
                with patch.object(useradd, "info", mock):
                    self.assertFalse(useradd.chother("salt", "foobar"))

        mock = MagicMock(return_value={"other": "foobar3"})
        with patch.object(useradd, "_get_gecos", mock):
            mock = MagicMock(return_value=None)
            with patch.dict(useradd.__salt__, {"cmd.run": mock}):
                mock = MagicMock(return_value={"other": "foobar3"})
                with patch.object(useradd, "info", mock):
                    self.assertFalse(useradd.chother("salt", "foobar"))

    # 'info' function tests: 1

    @skipIf(HAS_PWD is False, "The pwd module is not available")
    def test_info(self):
        """
        Test the user information
        """
        self.assertEqual(useradd.info("username-that-does-not-exist"), {})

        mock = MagicMock(
            return_value=pwd.struct_passwd(
                (
                    "_TEST_GROUP",
                    "*",
                    83,
                    83,
                    "AMaViS Daemon",
                    "/var/virusmails",
                    "/usr/bin/false",
                )
            )
        )
        with patch.object(pwd, "getpwnam", mock):
            self.assertEqual(
                useradd.info("username-that-does-not-exist")["name"], "_TEST_GROUP"
            )

    # 'list_groups' function tests: 1

    def test_list_groups(self):
        """
        Test if it return a list of groups the named user belongs to
        """
        with patch("salt.utils.user.get_group_list", MagicMock(return_value="Salt")):
            self.assertEqual(useradd.list_groups("name"), "Salt")

    # 'list_users' function tests: 1

    @skipIf(HAS_PWD is False, "The pwd module is not available")
    def test_list_users(self):
        """
        Test if it returns a list of all users
        """
        self.assertTrue(useradd.list_users())

    # 'list_users' function tests: 1

    def test_rename(self):
        """
        Test if the username for a named user changed
        """
        mock = MagicMock(return_value=False)
        with patch.object(useradd, "info", mock):
            self.assertRaises(CommandExecutionError, useradd.rename, "salt", 1)

        mock = MagicMock(return_value=True)
        with patch.object(useradd, "info", mock):
            self.assertRaises(CommandExecutionError, useradd.rename, "salt", 1)

        mock = MagicMock(return_value=None)
        with patch.dict(useradd.__salt__, {"cmd.run": mock}):
            mock = MagicMock(side_effect=[False, {"name": ""}, {"name": "salt"}])
            with patch.object(useradd, "info", mock):
                self.assertTrue(useradd.rename("name", "salt"))

        mock = MagicMock(return_value=None)
        with patch.dict(useradd.__salt__, {"cmd.run": mock}):
            mock = MagicMock(side_effect=[False, {"name": ""}, {"name": ""}])
            with patch.object(useradd, "info", mock):
                self.assertFalse(useradd.rename("salt", "salt"))

    def test_build_gecos_field(self):
        """
        Test if gecos fields are built correctly (removing trailing commas)
        """
        test_gecos = {
            "fullname": "Testing",
            "roomnumber": 1234,
            "workphone": 22222,
            "homephone": 99999,
        }
        expected_gecos_fields = "Testing,1234,22222,99999"
        self.assertEqual(useradd._build_gecos(test_gecos), expected_gecos_fields)
        test_gecos.pop("roomnumber")
        test_gecos.pop("workphone")
        expected_gecos_fields = "Testing,,,99999"
        self.assertEqual(useradd._build_gecos(test_gecos), expected_gecos_fields)
        test_gecos.pop("homephone")
        expected_gecos_fields = "Testing"
        self.assertEqual(useradd._build_gecos(test_gecos), expected_gecos_fields)



"""
Short-URL redirects
"""
import json
import os

import sphinx.ext.intersphinx

DOCS_URL = "https://docs.saltproject.io/en/latest/"


def write_urls_index(app, exc):
    """
    Generate a JSON file to serve as an index for short-URL lookups
    """
    inventory = os.path.join(app.builder.outdir, "objects.inv")
    objects = sphinx.ext.intersphinx.fetch_inventory(app, DOCS_URL, inventory)

    with open(os.path.join(app.builder.outdir, "shorturls.json"), "w") as f:
        json.dump(objects, f)


def setup(app):
    app.connect("build-finished", write_urls_index)



"""
Integration tests for the beacon states
"""
import pytest
from tests.support.case import ModuleCase
from tests.support.mixins import SaltReturnAssertsMixin


@pytest.mark.usefixtures("salt_sub_minion")
class BeaconStateTestCase(ModuleCase, SaltReturnAssertsMixin):
    """
    Test beacon states
    """

    def setUp(self):
        """ """
        self.run_function("beacons.reset", f_timeout=300)
        self.wait_for_all_jobs()

    def tearDown(self):
        self.run_function("beacons.reset", f_timeout=300)

    @pytest.mark.slow_test
    def test_present_absent(self):
        kwargs = {"/": "38%", "interval": 5}
        ret = self.run_state(
            "beacon.present", name="diskusage", f_timeout=300, **kwargs
        )
        self.assertSaltTrueReturn(ret)

        ret = self.run_function("beacons.list", return_yaml=False, f_timeout=300)
        self.assertTrue("diskusage" in ret)
        self.assertTrue({"interval": 5} in ret["diskusage"])
        self.assertTrue({"/": "38%"} in ret["diskusage"])

        ret = self.run_state("beacon.absent", name="diskusage", f_timeout=300)
        self.assertSaltTrueReturn(ret)

        ret = self.run_function("beacons.list", return_yaml=False, f_timeout=300)
        self.assertEqual(ret, {"beacons": {}})



import pytest
import salt.modules.postgres as postgresmod
import salt.states.postgres_extension as postgres_extension
from tests.support.mock import Mock, patch


@pytest.fixture
def configure_loader_modules():
    return {
        postgres_extension: {
            "__grains__": {"os_family": "linux"},
            "__salt__": {
                "config.option": Mock(),
                "cmd.run_all": Mock(),
                "file.chown": Mock(),
                "file.remove": Mock(),
            },
            "__opts__": {"test": False},
        },
    }


def test_present_failed():
    """
    scenario of creating upgrading extensions with possible schema and
    version specifications
    """
    with patch.dict(
        postgres_extension.__salt__,
        {
            "postgres.create_metadata": Mock(
                side_effect=[
                    [postgresmod._EXTENSION_NOT_INSTALLED],
                    [postgresmod._EXTENSION_TO_MOVE, postgresmod._EXTENSION_INSTALLED],
                ]
            ),
            "postgres.create_extension": Mock(side_effect=[False, False]),
        },
    ):
        ret = postgres_extension.present("foo")
        assert ret == {
            "comment": "Failed to install extension foo",
            "changes": {},
            "name": "foo",
            "result": False,
        }
        ret = postgres_extension.present("foo")
        assert ret == {
            "comment": "Failed to upgrade extension foo",
            "changes": {},
            "name": "foo",
            "result": False,
        }


def test_present():
    """
    scenario of creating upgrading extensions with possible schema and
    version specifications
    """
    with patch.dict(
        postgres_extension.__salt__,
        {
            "postgres.create_metadata": Mock(
                side_effect=[
                    [postgresmod._EXTENSION_NOT_INSTALLED],
                    [postgresmod._EXTENSION_INSTALLED],
                    [postgresmod._EXTENSION_TO_MOVE, postgresmod._EXTENSION_INSTALLED],
                ]
            ),
            "postgres.create_extension": Mock(side_effect=[True, True, True]),
        },
    ):
        ret = postgres_extension.present("foo")
        assert ret == {
            "comment": "The extension foo has been installed",
            "changes": {"foo": "Installed"},
            "name": "foo",
            "result": True,
        }
        ret = postgres_extension.present("foo")
        assert ret == {
            "comment": "Extension foo is already present",
            "changes": {},
            "name": "foo",
            "result": True,
        }
        ret = postgres_extension.present("foo")
        assert ret == {
            "comment": "The extension foo has been upgraded",
            "changes": {"foo": "Upgraded"},
            "name": "foo",
            "result": True,
        }


def test_presenttest():
    """
    scenario of creating upgrading extensions with possible schema and
    version specifications
    """
    with patch.dict(
        postgres_extension.__salt__,
        {
            "postgres.create_metadata": Mock(
                side_effect=[
                    [postgresmod._EXTENSION_NOT_INSTALLED],
                    [postgresmod._EXTENSION_INSTALLED],
                    [postgresmod._EXTENSION_TO_MOVE, postgresmod._EXTENSION_INSTALLED],
                ]
            ),
            "postgres.create_extension": Mock(side_effect=[True, True, True]),
        },
    ):
        with patch.dict(postgres_extension.__opts__, {"test": True}):
            ret = postgres_extension.present("foo")
            assert ret == {
                "comment": "Extension foo is set to be installed",
                "changes": {},
                "name": "foo",
                "result": None,
            }
            ret = postgres_extension.present("foo")
            assert ret == {
                "comment": "Extension foo is already present",
                "changes": {},
                "name": "foo",
                "result": True,
            }
            ret = postgres_extension.present("foo")
            assert ret == {
                "comment": "Extension foo is set to be upgraded",
                "changes": {},
                "name": "foo",
                "result": None,
            }


def test_absent():
    """
    scenario of creating upgrading extensions with possible schema and
    version specifications
    """
    with patch.dict(
        postgres_extension.__salt__,
        {
            "postgres.is_installed_extension": Mock(side_effect=[True, False]),
            "postgres.drop_extension": Mock(side_effect=[True, True]),
        },
    ):
        ret = postgres_extension.absent("foo")
        assert ret == {
            "comment": "Extension foo has been removed",
            "changes": {"foo": "Absent"},
            "name": "foo",
            "result": True,
        }
        ret = postgres_extension.absent("foo")
        assert ret == {
            "comment": "Extension foo is not present, so it cannot be removed",
            "changes": {},
            "name": "foo",
            "result": True,
        }


def test_absent_failed():
    """
    scenario of creating upgrading extensions with possible schema and
    version specifications
    """
    with patch.dict(postgres_extension.__opts__, {"test": False}):
        with patch.dict(
            postgres_extension.__salt__,
            {
                "postgres.is_installed_extension": Mock(side_effect=[True, True]),
                "postgres.drop_extension": Mock(side_effect=[False, False]),
            },
        ):
            ret = postgres_extension.absent("foo")
            assert ret == {
                "comment": "Extension foo failed to be removed",
                "changes": {},
                "name": "foo",
                "result": False,
            }


def test_absent_failedtest():
    with patch.dict(
        postgres_extension.__salt__,
        {
            "postgres.is_installed_extension": Mock(side_effect=[True, True]),
            "postgres.drop_extension": Mock(side_effect=[False, False]),
        },
    ):
        with patch.dict(postgres_extension.__opts__, {"test": True}):
            ret = postgres_extension.absent("foo")
        assert ret == {
            "comment": "Extension foo is set to be removed",
            "changes": {},
            "name": "foo",
            "result": None,
        }



"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>
"""

import os

import salt.modules.nagios as nagios
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class NagiosTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.nagios
    """

    def setup_loader_modules(self):
        return {nagios: {}}

    def test_run(self):
        """
        Test for Run nagios plugin and return all
         the data execution with cmd.run
        """
        with patch.object(nagios, "_execute_cmd", return_value="A"):
            self.assertEqual(nagios.run("plugin"), "A")

    def test_retcode(self):
        """
        Test for Run one nagios plugin and return retcode of the execution
        """
        with patch.object(nagios, "_execute_cmd", return_value="A"):
            self.assertEqual(
                nagios.retcode("plugin", key_name="key"), {"key": {"status": "A"}}
            )

    def test_run_all(self):
        """
        Test for Run nagios plugin and return all
         the data execution with cmd.run_all
        """
        with patch.object(nagios, "_execute_cmd", return_value="A"):
            self.assertEqual(nagios.run_all("plugin"), "A")

    def test_retcode_pillar(self):
        """
        Test for Run one or more nagios plugins from pillar data and
         get the result of cmd.retcode
        """
        with patch.dict(nagios.__salt__, {"pillar.get": MagicMock(return_value={})}):
            self.assertEqual(nagios.retcode_pillar("pillar_name"), {})

    def test_run_pillar(self):
        """
        Test for Run one or more nagios plugins from pillar data
         and get the result of cmd.run
        """
        with patch.object(nagios, "_execute_pillar", return_value="A"):
            self.assertEqual(nagios.run_pillar("pillar"), "A")

    def test_run_all_pillar(self):
        """
        Test for Run one or more nagios plugins from pillar data
         and get the result of cmd.run
        """
        with patch.object(nagios, "_execute_pillar", return_value="A"):
            self.assertEqual(nagios.run_all_pillar("pillar"), "A")

    def test_list_plugins(self):
        """
        Test for List all the nagios plugins
        """
        with patch.object(os, "listdir", return_value=[]):
            self.assertEqual(nagios.list_plugins(), [])



"""
    :codeauthor: Florian Kantelberg <florian.kantelberg@initos.com>
"""

import pytest
from salt.cloud.clouds import hetzner
from salt.exceptions import SaltCloudException, SaltCloudSystemExit
from tests.support.mock import MagicMock, patch


class ModelMock(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model = MagicMock()
        self.model.__slots__ = list(self)

        for attr, value in self.items():
            setattr(self, attr, value)


@pytest.fixture
def configure_loader_modules():
    return {
        hetzner: {
            "__active_provider_name__": "",
            "__utils__": {
                "cloud.bootstrap": MagicMock(),
                "cloud.cache_node": MagicMock(),
                "cloud.delete_minion_cachedir": MagicMock(),
                "cloud.filter_event": MagicMock(),
                "cloud.fire_event": MagicMock(),
            },
            "__opts__": {
                "providers": {
                    "hetzner1": {"hetzner": {"driver": "hetzner", "key": "abcdefg"}},
                },
                "profiles": {},
                "sock_dir": "/var/sockxxx",
                "transport": "tcp",
                "update_cachedir": True,
            },
        },
    }


@pytest.fixture
def images():
    return {"ubuntu-20.04": {"name": "ubuntu-20.04", "id": 15512617}}


@pytest.fixture
def locations():
    return {"fsn1": {"city": "Falkenstein", "name": "fsn1", "id": 1}}


@pytest.fixture
def sizes():
    return {"cpx21": {"name": "cpx21", "cores": 3, "id": 32}}


@pytest.fixture
def ssh_keys():
    return {"myssh": {"name": "myssh"}}


@pytest.fixture
def vm():
    return {
        "name": "myvm",
        "driver": "hetzner",
        "size": "cpx21",
        "image": "ubuntu-20.04",
    }


def test_virtual():
    with patch("salt.cloud.clouds.hetzner.get_configured_provider") as config:
        with patch("salt.cloud.clouds.hetzner.get_dependencies") as dependency:
            config.return_value = False
            dependency.return_value = False
            assert not hetzner.__virtual__()

            config.return_value = True
            assert not hetzner.__virtual__()

            dependency.return_value = True
            assert hetzner.__virtual__() == hetzner.__virtualname__


def test_object_to_dict():
    mock = MagicMock()
    mock.attr1 = "abc"
    mock.attr2 = "def"

    assert hetzner._object_to_dict(mock, ["attr1", "attr2"]) == {
        "attr1": "abc",
        "attr2": "def",
    }


def test_dependencies():
    with patch("salt.cloud.clouds.hetzner.HAS_HCLOUD", True):
        assert hetzner.get_dependencies()
    with patch("salt.cloud.clouds.hetzner.HAS_HCLOUD", False):
        assert not hetzner.get_dependencies()


@pytest.mark.skipif(
    not hetzner.HAS_HCLOUD, reason="Install hcloud to be able to run this test."
)
def test_connect_client():
    with patch("salt.cloud.clouds.hetzner.hcloud"):
        hetzner._connect_client()
        hetzner.hcloud.Client.assert_called_once_with("abcdefgh")


def test_avail_images_action(images):
    with patch(
        "salt.cloud.clouds.hetzner._connect_client", return_value=MagicMock()
    ) as connect:
        with pytest.raises(SaltCloudSystemExit):
            hetzner.avail_images("action")

        connect.return_value.images.get_all.return_value = map(
            ModelMock, images.values()
        )
        assert hetzner.avail_images() == images


def test_avail_locations(locations):
    with patch(
        "salt.cloud.clouds.hetzner._connect_client", return_value=MagicMock()
    ) as connect:
        with pytest.raises(SaltCloudSystemExit):
            hetzner.avail_locations("action")

        connect.return_value.locations.get_all.return_value = map(
            ModelMock, locations.values()
        )
        assert hetzner.avail_locations() == locations


def test_avail_sizes(sizes):
    with patch(
        "salt.cloud.clouds.hetzner._connect_client", return_value=MagicMock()
    ) as connect:
        with pytest.raises(SaltCloudSystemExit):
            hetzner.avail_sizes("action")

        connect.return_value.server_types.get_all.return_value = map(
            ModelMock, sizes.values()
        )
        assert hetzner.avail_sizes() == sizes


def test_list_ssh_keys(ssh_keys):
    with patch(
        "salt.cloud.clouds.hetzner._connect_client", return_value=MagicMock()
    ) as connect:
        with pytest.raises(SaltCloudSystemExit):
            hetzner.list_ssh_keys("action")

        connect.return_value.ssh_keys.get_all.return_value = map(
            ModelMock, ssh_keys.values()
        )
        assert hetzner.list_ssh_keys() == ssh_keys


def test_list_nodes_full():
    """Test the list_nodes_full function by using a mock"""
    with patch(
        "salt.cloud.clouds.hetzner._connect_client", return_value=MagicMock()
    ) as connect:
        with pytest.raises(SaltCloudSystemExit):
            hetzner.list_nodes_full("action")

        mock = MagicMock()
        mock.id = 123456
        mock.name = "abc"
        mock.public_net.ipv4.ip = "127.0.0.1/32"
        mock.public_net.ipv6.ip = "::1/64"

        private_net_mock = MagicMock()
        private_net_mock.ip = "10.0.0.1/16"
        mock.private_net = []
        mock.private_net.append(private_net_mock)

        mock.labels = "abc"
        connect.return_value.servers.get_all.return_value = [mock]

        nodes = hetzner.list_nodes_full()
        assert nodes[mock.name]["id"], mock.id
        # Labels shouldn't be filtered
        assert "labels" in nodes[mock.name]

        assert nodes[mock.name]["public_ips"]["ipv4"] == "127.0.0.1/32"
        assert nodes[mock.name]["public_ips"]["ipv6"] == "::1/64"

        assert nodes[mock.name]["private_ips"][0]["ip"] == "10.0.0.1/16"


def test_list_nodes():
    """Test the list_nodes function by using a mock"""
    with patch(
        "salt.cloud.clouds.hetzner._connect_client", return_value=MagicMock()
    ) as connect:
        with pytest.raises(SaltCloudSystemExit):
            hetzner.list_nodes("action")

        mock = MagicMock()
        mock.id = 123456
        mock.name = "abc"
        mock.public_net.ipv4.ip = "127.0.0.1/32"
        mock.public_net.ipv6.ip = "::1/64"

        private_net_mock = MagicMock()
        private_net_mock.ip = "10.0.0.1/16"
        mock.private_net = []
        mock.private_net.append(private_net_mock)

        mock.labels = "abc"
        connect.return_value.servers.get_all.return_value = [mock]

        nodes = hetzner.list_nodes()
        assert nodes[mock.name]["id"], mock.id
        # Labels should be filtered
        assert "labels" not in nodes[mock.name]

        assert nodes[mock.name]["public_ips"]["ipv4"] == "127.0.0.1/32"
        assert nodes[mock.name]["public_ips"]["ipv6"] == "::1/64"

        assert nodes[mock.name]["private_ips"][0]["ip"] == "10.0.0.1/16"


def test_show_instance():
    """Test the show_instance function by using a mock"""
    with patch(
        "salt.cloud.clouds.hetzner._connect_client", return_value=MagicMock()
    ) as connect:
        with pytest.raises(SaltCloudSystemExit):
            hetzner.show_instance("myvm", "action")

        mock = MagicMock()
        mock.id = 123456
        mock.name = "abc"
        mock.public_net.ipv4.ip = "127.0.0.1/32"
        mock.public_net.ipv6.ip = "::1/64"
        mock.private_net = []
        mock.labels = "abc"
        connect.return_value.servers.get_all.return_value = [mock]

        nodes = hetzner.show_instance(mock.name)
        assert nodes["id"] == mock.id

        nodes = hetzner.show_instance("not-existing")
        assert nodes == {}


def test_wait_until():
    """Test the wait_until function"""

    with patch("salt.cloud.clouds.hetzner.show_instance") as show_instance:
        show_instance.side_effect = [{"state": "done"}, IndexError()]
        assert hetzner.wait_until("abc", "done")

        show_instance.side_effect = [{"state": "done"}, IndexError()]
        with pytest.raises(IndexError):
            hetzner.wait_until("abc", "never")

        show_instance.side_effect = [{"state": "done"}, IndexError()]
        assert not hetzner.wait_until("abc", "never", timeout=0)


def test_create(images, sizes, vm):
    """Test the overall creation and the required parameters"""
    with patch(
        "salt.cloud.clouds.hetzner._connect_client", return_value=MagicMock()
    ) as connect:
        with patch("salt.cloud.clouds.hetzner.wait_until", return_value=True) as wait:
            with pytest.raises(SaltCloudException):
                hetzner.create({})

            connect.return_value.server_types.get_by_name.return_value = None
            connect.return_value.images.get_by_name.return_value = None
            with pytest.raises(SaltCloudException):
                hetzner.create(vm)

            connect.return_value.server_types.get_by_name.return_value = ModelMock(
                sizes["cpx21"]
            )
            with pytest.raises(SaltCloudException):
                hetzner.create(vm)

            connect.return_value.images.get_by_name.return_value = ModelMock(
                images["ubuntu-20.04"]
            )

            assert hetzner.create(vm)["created"]
            connect.return_value.servers.create.assert_called_once()
            args = connect.return_value.servers.create.call_args
            assert args.kwargs["name"] == vm["name"]
            assert args.kwargs["server_type"] == sizes["cpx21"]
            assert args.kwargs["image"] == images["ubuntu-20.04"]


def test_create_location(vm):
    """Test the locations during the creation"""
    vm["location"] = "abc"

    with patch(
        "salt.cloud.clouds.hetzner._connect_client", return_value=MagicMock()
    ) as connect:
        hetzner.create(vm)
        connect.return_value.servers.create.assert_called_once()
        connect.return_value.locations.get_by_name.assert_called_once_with("abc")

        # The location should be passed to the create
        connect.return_value.locations.get_by_name.return_value = "abc"
        hetzner.create(vm)

        args = connect.return_value.servers.create.call_args
        assert args.kwargs["location"] == "abc"

        # Stop if the location is invalid
        connect.return_value.locations.get_by_name.return_value = None
        with pytest.raises(SaltCloudException):
            hetzner.create(vm)


def test_ssh_keys(vm):
    """Test the locations during the creation"""
    vm["ssh_keys"] = ["me"]

    with patch(
        "salt.cloud.clouds.hetzner._connect_client", return_value=MagicMock()
    ) as connect:
        hetzner.create(vm)
        connect.return_value.ssh_keys.get_by_name.assert_called_once_with("me")

        # The ssh_keys should be passed to the create
        connect.return_value.ssh_keys.get_by_name.return_value = "me"
        hetzner.create(vm)
        args = connect.return_value.servers.create.call_args
        assert args.kwargs["ssh_keys"] == ["me"]

        # Invalid keys should be sorted out
        connect.return_value.ssh_keys.get_by_name.return_value = None
        hetzner.create(vm)
        args = connect.return_value.servers.create.call_args
        assert args.kwargs["ssh_keys"] == []


def test_create_datacenter(vm):
    """Test the datacenters during the creation"""
    vm["datacenter"] = "abc"

    with patch(
        "salt.cloud.clouds.hetzner._connect_client", return_value=MagicMock()
    ) as connect:
        hetzner.create(vm)
        connect.return_value.servers.create.assert_called_once()
        connect.return_value.datacenters.get_by_name.assert_called_once_with("abc")

        # The datacenter should be passed to the create
        connect.return_value.datacenters.get_by_name.return_value = "abc"
        hetzner.create(vm)
        args = connect.return_value.servers.create.call_args
        assert args.kwargs["datacenter"] == "abc"

        # Stop if the datacenter is invalid
        connect.return_value.datacenters.get_by_name.return_value = None
        with pytest.raises(SaltCloudException):
            hetzner.create(vm)


def test_create_volumes(vm):
    """Test the volumes during the creation"""
    vm["volumes"] = ["a", "b"]

    with patch(
        "salt.cloud.clouds.hetzner._connect_client", return_value=MagicMock()
    ) as connect:
        connect.return_value.volumes.get_all.return_value = ["a", "c"]

        hetzner.create(vm)
        args = connect.return_value.servers.create.call_args
        assert args.kwargs["volumes"], ["a"]

        vm["volumes"] = ["a", "b", "c"]
        hetzner.create(vm)
        args = connect.return_value.servers.create.call_args
        assert args.kwargs["volumes"] == ["a", "c"]


def test_create_networks(vm):
    """Test the networks during the creation"""
    vm["networks"] = ["a", "b"]

    with patch(
        "salt.cloud.clouds.hetzner._connect_client", return_value=MagicMock()
    ) as connect:
        connect.return_value.networks.get_all.return_value = ["a", "c"]

        hetzner.create(vm)
        args = connect.return_value.servers.create.call_args
        assert args.kwargs["networks"] == ["a"]

        vm["networks"] = ["a", "b", "c"]
        hetzner.create(vm)
        args = connect.return_value.servers.create.call_args
        assert args.kwargs["networks"] == ["a", "c"]


def test_start():
    """Test the start action"""
    with patch(
        "salt.cloud.clouds.hetzner._connect_client", return_value=MagicMock()
    ) as connect:
        with patch("salt.cloud.clouds.hetzner.wait_until", return_value=True) as wait:
            with pytest.raises(SaltCloudSystemExit):
                hetzner.start("myvm", "function")

            connect.return_value.servers.get_by_name.return_value = None
            hetzner.start("myvm", "action")

            server = connect.return_value.servers.get_by_name.return_value = MagicMock()

            assert "Started" in hetzner.start("myvm", "action")
            server.power_on.assert_called_once()

            wait.return_value = False
            hetzner.start("myvm", "action")


def test_stop():
    """Test the stop action"""
    with patch(
        "salt.cloud.clouds.hetzner._connect_client", return_value=MagicMock()
    ) as connect:
        with patch("salt.cloud.clouds.hetzner.wait_until", return_value=True) as wait:
            with pytest.raises(SaltCloudSystemExit):
                hetzner.stop("myvm", "function")

            connect.return_value.servers.get_by_name.return_value = None
            hetzner.stop("myvm", "action")

            server = connect.return_value.servers.get_by_name.return_value = MagicMock()

            assert "Stopped" in hetzner.stop("myvm", "action")
            server.power_off.assert_called_once()

            wait.return_value = False
            hetzner.stop("myvm", "action")


def test_reboot():
    """Test the reboot action"""
    with patch(
        "salt.cloud.clouds.hetzner._connect_client", return_value=MagicMock()
    ) as connect:
        with patch("salt.cloud.clouds.hetzner.wait_until", return_value=True) as wait:
            with pytest.raises(SaltCloudSystemExit):
                hetzner.reboot("myvm", "function")

            connect.return_value.servers.get_by_name.return_value = None
            hetzner.reboot("myvm", "action")

            server = connect.return_value.servers.get_by_name.return_value = MagicMock()

            assert "Rebooted" in hetzner.reboot("myvm", "action")
            server.reboot.assert_called_once()

            wait.return_value = False
            hetzner.reboot("myvm", "action")


def test_destroy():
    """Test the destroy action"""
    with patch(
        "salt.cloud.clouds.hetzner._connect_client", return_value=MagicMock()
    ) as connect:
        with patch("salt.cloud.clouds.hetzner.wait_until", return_value=True) as wait:
            with patch("salt.cloud.clouds.hetzner.show_instance") as show_instance:
                with pytest.raises(SaltCloudSystemExit):
                    hetzner.destroy("myvm", "function")

                wait.return_value = False
                show_instance.return_value = {"state": "running"}
                connect.return_value.servers.get_by_name.return_value = None
                hetzner.destroy("myvm", "action")

                server = (
                    connect.return_value.servers.get_by_name.return_value
                ) = MagicMock()

                # Stop the server before shutdown but failed
                hetzner.destroy("myvm", "action")
                server.delete.assert_not_called()
                wait.assert_called_once_with("myvm", "off")

                wait.return_value = True
                hetzner.destroy("myvm", "action")
                server.delete.assert_called_once()

                # Don't stop if the server isn't running
                show_instance.return_value = {"state": "off"}
                wait.reset_mock()
                hetzner.destroy("myvm", "action")
                wait.assert_not_called()


def test_resize():
    """Test the resize action"""
    kwargs = {"size": "cpx21"}

    with patch(
        "salt.cloud.clouds.hetzner._connect_client", return_value=MagicMock()
    ) as connect:
        with patch("salt.cloud.clouds.hetzner.wait_until", return_value=True) as wait:
            with patch("salt.cloud.clouds.hetzner.show_instance") as show_instance:
                with pytest.raises(SaltCloudSystemExit):
                    hetzner.resize("myvm", kwargs, "function")

                wait.return_value = False
                show_instance.return_value = {"state": "running"}
                connect.return_value.server_types.get_by_name = MagicMock(
                    return_value=None
                )
                connect.return_value.servers.get_by_name.return_value = None

                hetzner.resize("myvm", kwargs, "action")

                server = (
                    connect.return_value.servers.get_by_name.return_value
                ) = MagicMock()

                # Invalid server size
                with pytest.raises(SaltCloudException):
                    hetzner.resize("myvm", {}, "action")
                with pytest.raises(SaltCloudException):
                    hetzner.resize("myvm", kwargs, "action")

                connect.return_value.server_types.get_by_name.return_value = True

                # Stop the server before shutdown but failed
                hetzner.resize("myvm", kwargs, "action")
                server.change_type.assert_not_called()
                wait.assert_called_once_with("myvm", "off")

                wait.return_value = True
                hetzner.resize("myvm", kwargs, "action")
                server.change_type.assert_called_once()

                # Don't stop if the server isn't running
                show_instance.return_value = {"state": "off"}
                wait.reset_mock()
                hetzner.resize("myvm", kwargs, "action")
                wait.assert_not_called()



import os

import pytest
import salt.modules.tls as tls
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {tls: {}}


@pytest.fixture(scope="module")
def tls_test_data():
    return {
        "create_ca": {
            "bits": 2048,
            "CN": "localhost",
            "C": "US",
            "ST": "Utah",
            "L": "Salt Lake City",
            "O": "SaltStack",
            "OU": "Test Unit",
            "emailAddress": "xyz@pdq.net",
            "digest": "sha256",
            "replace": False,
        }
    }


@pytest.mark.skip_on_windows(reason="Skipping on Windows per Shane's suggestion")
def test_create_ca_permissions_on_cert_and_key(tmpdir, tls_test_data):
    ca_name = "test_ca"
    certp = tmpdir.join(ca_name).join("{}_ca_cert.crt".format(ca_name)).strpath
    certk = tmpdir.join(ca_name).join("{}_ca_cert.key".format(ca_name)).strpath
    mock_opt = MagicMock(return_value=tmpdir)
    mock_ret = MagicMock(return_value=0)

    with patch.dict(
        tls.__salt__, {"config.option": mock_opt, "cmd.retcode": mock_ret}
    ), patch.dict(tls.__opts__, {"hash_type": "sha256", "cachedir": str(tmpdir)}):
        tls.create_ca(ca_name, days=365, fixmode=False, **tls_test_data["create_ca"])
        certp_mode = os.stat(certp).st_mode & 0o7777
        certk_mode = os.stat(certk).st_mode & 0o7777

        assert 0o644 == certp_mode
        assert 0o600 == certk_mode


@pytest.mark.skip_on_windows(reason="Skipping on Windows per Shane's suggestion")
def test_create_csr_permissions_on_csr_and_key(tmpdir, tls_test_data):
    ca_name = "test_ca"
    csrp = (
        tmpdir.join(ca_name)
        .join("certs")
        .join("{}.csr".format(tls_test_data["create_ca"]["CN"]))
        .strpath
    )
    keyp = (
        tmpdir.join(ca_name)
        .join("certs")
        .join("{}.key".format(tls_test_data["create_ca"]["CN"]))
        .strpath
    )

    mock_opt = MagicMock(return_value=tmpdir)
    mock_ret = MagicMock(return_value=0)
    mock_pgt = MagicMock(return_value=False)

    with patch.dict(
        tls.__salt__,
        {"config.option": mock_opt, "cmd.retcode": mock_ret, "pillar.get": mock_pgt},
    ), patch.dict(tls.__opts__, {"hash_type": "sha256", "cachedir": str(tmpdir)}):
        tls.create_ca(ca_name, days=365, **tls_test_data["create_ca"])
        tls.create_csr(ca_name, **tls_test_data["create_ca"])

        csrp_mode = os.stat(csrp).st_mode & 0o7777
        keyp_mode = os.stat(keyp).st_mode & 0o7777

        assert 0o644 == csrp_mode
        assert 0o600 == keyp_mode


@pytest.mark.skip_on_windows(reason="Skipping on Windows per Shane's suggestion")
def test_create_self_signed_cert_permissions_on_csr_cert_and_key(tmpdir, tls_test_data):
    ca_name = "test_ca"
    certp = (
        tmpdir.join(ca_name)
        .join("certs")
        .join("{}.crt".format(tls_test_data["create_ca"]["CN"]))
        .strpath
    )
    keyp = (
        tmpdir.join(ca_name)
        .join("certs")
        .join("{}.key".format(tls_test_data["create_ca"]["CN"]))
        .strpath
    )

    mock_opt = MagicMock(return_value=tmpdir)
    mock_ret = MagicMock(return_value=0)
    mock_pgt = MagicMock(return_value=False)

    with patch.dict(
        tls.__salt__,
        {"config.option": mock_opt, "cmd.retcode": mock_ret, "pillar.get": mock_pgt},
    ), patch.dict(tls.__opts__, {"hash_type": "sha256", "cachedir": str(tmpdir)}):
        tls.create_self_signed_cert(ca_name, days=365, **tls_test_data["create_ca"])

        certp_mode = os.stat(certp).st_mode & 0o7777
        keyp_mode = os.stat(keyp).st_mode & 0o7777

        assert 0o644 == certp_mode
        assert 0o600 == keyp_mode



"""
Module for managing Solaris logadm based log rotations.
"""

import logging
import shlex

import salt.utils.args
import salt.utils.decorators as decorators
import salt.utils.files
import salt.utils.stringutils

try:
    from shlex import quote as _quote_args  # pylint: disable=E0611
except ImportError:
    from pipes import quote as _quote_args


log = logging.getLogger(__name__)
default_conf = "/etc/logadm.conf"
option_toggles = {
    "-c": "copy",
    "-l": "localtime",
    "-N": "skip_missing",
}
option_flags = {
    "-A": "age",
    "-C": "count",
    "-a": "post_command",
    "-b": "pre_command",
    "-e": "mail_addr",
    "-E": "expire_command",
    "-g": "group",
    "-m": "mode",
    "-M": "rename_command",
    "-o": "owner",
    "-p": "period",
    "-P": "timestmp",
    "-R": "old_created_command",
    "-s": "size",
    "-S": "max_size",
    "-t": "template",
    "-T": "old_pattern",
    "-w": "entryname",
    "-z": "compress_count",
}


def __virtual__():
    """
    Only work on Solaris based systems
    """
    if "Solaris" in __grains__["os_family"]:
        return True
    return (
        False,
        "The logadm execution module cannot be loaded: only available on Solaris.",
    )


def _arg2opt(arg):
    """
    Turn a pass argument into the correct option
    """
    res = [o for o, a in option_toggles.items() if a == arg]
    res += [o for o, a in option_flags.items() if a == arg]
    return res[0] if res else None


def _parse_conf(conf_file=default_conf):
    """
    Parse a logadm configuration file.
    """
    ret = {}
    with salt.utils.files.fopen(conf_file, "r") as ifile:
        for line in ifile:
            line = salt.utils.stringutils.to_unicode(line).strip()
            if not line:
                continue
            if line.startswith("#"):
                continue
            splitline = line.split(" ", 1)
            ret[splitline[0]] = splitline[1]
    return ret


def _parse_options(entry, options, include_unset=True):
    """
    Parse a logadm options string
    """
    log_cfg = {}
    options = shlex.split(options)
    if not options:
        return None

    ## identifier is entry or log?
    if entry.startswith("/"):
        log_cfg["log_file"] = entry
    else:
        log_cfg["entryname"] = entry

    ## parse options
    # NOTE: we loop over the options because values may exist multiple times
    index = 0
    while index < len(options):
        # log file
        if index in [0, (len(options) - 1)] and options[index].startswith("/"):
            log_cfg["log_file"] = options[index]

        # check if toggle option
        elif options[index] in option_toggles:
            log_cfg[option_toggles[options[index]]] = True

        # check if flag option
        elif options[index] in option_flags and (index + 1) <= len(options):
            log_cfg[option_flags[options[index]]] = (
                int(options[index + 1])
                if options[index + 1].isdigit()
                else options[index + 1]
            )
            index += 1

        # unknown options
        else:
            if "additional_options" not in log_cfg:
                log_cfg["additional_options"] = []
            if " " in options[index]:
                log_cfg["dditional_options"] = "'{}'".format(options[index])
            else:
                log_cfg["additional_options"].append(options[index])

        index += 1

    ## turn additional_options into string
    if "additional_options" in log_cfg:
        log_cfg["additional_options"] = " ".join(log_cfg["additional_options"])

    ## ensure we have a log_file
    # NOTE: logadm assumes logname is a file if no log_file is given
    if "log_file" not in log_cfg and "entryname" in log_cfg:
        log_cfg["log_file"] = log_cfg["entryname"]
        del log_cfg["entryname"]

    ## include unset
    if include_unset:
        # toggle optioons
        for name in option_toggles.values():
            if name not in log_cfg:
                log_cfg[name] = False

        # flag options
        for name in option_flags.values():
            if name not in log_cfg:
                log_cfg[name] = None

    return log_cfg


def show_conf(conf_file=default_conf, name=None):
    """
    Show configuration

    conf_file : string
        path to logadm.conf, defaults to /etc/logadm.conf
    name : string
        optional show only a single entry

    CLI Example:

    .. code-block:: bash

        salt '*' logadm.show_conf
        salt '*' logadm.show_conf name=/var/log/syslog
    """
    cfg = _parse_conf(conf_file)

    # filter
    if name and name in cfg:
        return {name: cfg[name]}
    elif name:
        return {name: "not found in {}".format(conf_file)}
    else:
        return cfg


def list_conf(conf_file=default_conf, log_file=None, include_unset=False):
    """
    Show parsed configuration

    .. versionadded:: 2018.3.0

    conf_file : string
        path to logadm.conf, defaults to /etc/logadm.conf
    log_file : string
        optional show only one log file
    include_unset : boolean
        include unset flags in output

    CLI Example:

    .. code-block:: bash

        salt '*' logadm.list_conf
        salt '*' logadm.list_conf log=/var/log/syslog
        salt '*' logadm.list_conf include_unset=False
    """
    cfg = _parse_conf(conf_file)
    cfg_parsed = {}

    ## parse all options
    for entry in cfg:
        log_cfg = _parse_options(entry, cfg[entry], include_unset)
        cfg_parsed[
            log_cfg["log_file"] if "log_file" in log_cfg else log_cfg["entryname"]
        ] = log_cfg

    ## filter
    if log_file and log_file in cfg_parsed:
        return {log_file: cfg_parsed[log_file]}
    elif log_file:
        return {log_file: "not found in {}".format(conf_file)}
    else:
        return cfg_parsed


@decorators.memoize
def show_args():
    """
    Show which arguments map to which flags and options.

    .. versionadded:: 2018.3.0

    CLI Example:

    .. code-block:: bash

        salt '*' logadm.show_args
    """
    mapping = {"flags": {}, "options": {}}
    for flag, arg in option_toggles.items():
        mapping["flags"][flag] = arg
    for option, arg in option_flags.items():
        mapping["options"][option] = arg

    return mapping


def rotate(name, pattern=None, conf_file=default_conf, **kwargs):
    """
    Set up pattern for logging.

    name : string
        alias for entryname
    pattern : string
        alias for log_file
    conf_file : string
        optional path to alternative configuration file
    kwargs : boolean|string|int
        optional additional flags and parameters

    .. note::
        ``name`` and ``pattern`` were kept for backwards compatibility reasons.

        ``name`` is an alias for the ``entryname`` argument, ``pattern`` is an alias
        for ``log_file``. These aliases will only be used if the ``entryname`` and
        ``log_file`` arguments are not passed.

        For a full list of arguments see ```logadm.show_args```.

    CLI Example:

    .. code-block:: bash

        salt '*' logadm.rotate myapplog pattern='/var/log/myapp/*.log' count=7
        salt '*' logadm.rotate myapplog log_file='/var/log/myapp/*.log' count=4 owner=myappd mode='0700'

    """
    ## cleanup kwargs
    kwargs = salt.utils.args.clean_kwargs(**kwargs)

    ## inject name into kwargs
    if "entryname" not in kwargs and name and not name.startswith("/"):
        kwargs["entryname"] = name

    ## inject pattern into kwargs
    if "log_file" not in kwargs:
        if pattern and pattern.startswith("/"):
            kwargs["log_file"] = pattern
        # NOTE: for backwards compatibility check if name is a path
        elif name and name.startswith("/"):
            kwargs["log_file"] = name

    ## build command
    log.debug("logadm.rotate - kwargs: %s", kwargs)
    command = "logadm -f {}".format(conf_file)
    for arg, val in kwargs.items():
        if arg in option_toggles.values() and val:
            command = "{} {}".format(
                command,
                _arg2opt(arg),
            )
        elif arg in option_flags.values():
            command = "{} {} {}".format(command, _arg2opt(arg), _quote_args(str(val)))
        elif arg != "log_file":
            log.warning("Unknown argument %s, don't know how to map this!", arg)
    if "log_file" in kwargs:
        # NOTE: except from ```man logadm```
        #   If no log file name is provided on a logadm command line, the entry
        #   name is assumed to be the same as the log file name. For example,
        #   the following two lines achieve the same thing, keeping two copies
        #   of rotated log files:
        #
        #     % logadm -C2 -w mylog /my/really/long/log/file/name
        #     % logadm -C2 -w /my/really/long/log/file/name
        if "entryname" not in kwargs:
            command = "{} -w {}".format(command, _quote_args(kwargs["log_file"]))
        else:
            command = "{} {}".format(command, _quote_args(kwargs["log_file"]))

    log.debug("logadm.rotate - command: %s", command)
    result = __salt__["cmd.run_all"](command, python_shell=False)
    if result["retcode"] != 0:
        return dict(Error="Failed in adding log", Output=result["stderr"])

    return dict(Result="Success")


def remove(name, conf_file=default_conf):
    """
    Remove log pattern from logadm

    CLI Example:

    .. code-block:: bash

      salt '*' logadm.remove myapplog
    """
    command = "logadm -f {} -r {}".format(conf_file, name)
    result = __salt__["cmd.run_all"](command, python_shell=False)
    if result["retcode"] != 0:
        return dict(
            Error="Failure in removing log. Possibly already removed?",
            Output=result["stderr"],
        )
    return dict(Result="Success")



import pytest

pytestmark = [
    pytest.mark.skip_on_windows(reason="salt-ssh not available on Windows"),
]


@pytest.fixture(scope="module")
def pillar_tree(base_env_pillar_tree_root_dir):
    top_file = """
    base:
      'localhost':
        - basic
      '127.0.0.1':
        - basic
    """
    basic_pillar_file = """
    monty: python
    knights:
      - Lancelot
      - Galahad
      - Bedevere
      - Robin
    """
    top_tempfile = pytest.helpers.temp_file(
        "top.sls", top_file, base_env_pillar_tree_root_dir
    )
    basic_tempfile = pytest.helpers.temp_file(
        "basic.sls", basic_pillar_file, base_env_pillar_tree_root_dir
    )

    with top_tempfile, basic_tempfile:
        yield


@pytest.mark.slow_test
def test_pillar_items(salt_ssh_cli, pillar_tree):
    """
    test pillar.items with salt-ssh
    """
    ret = salt_ssh_cli.run("pillar.items")
    assert ret.exitcode == 0
    assert ret.json
    pillar_items = ret.json
    assert "monty" in pillar_items
    assert pillar_items["monty"] == "python"
    assert "knights" in pillar_items
    assert pillar_items["knights"] == ["Lancelot", "Galahad", "Bedevere", "Robin"]


@pytest.mark.slow_test
def test_pillar_get(salt_ssh_cli, pillar_tree):
    """
    test pillar.get with salt-ssh
    """
    ret = salt_ssh_cli.run("pillar.get", "monty")
    assert ret.exitcode == 0
    assert ret.json
    assert ret.json == "python"


@pytest.mark.slow_test
def test_pillar_get_doesnotexist(salt_ssh_cli, pillar_tree):
    """
    test pillar.get when pillar does not exist with salt-ssh
    """
    ret = salt_ssh_cli.run("pillar.get", "doesnotexist")
    assert ret.exitcode == 0
    assert ret.json == ""



r"""
Renderer that will decrypt NACL ciphers

Any key in the SLS file can be an NACL cipher, and this renderer will decrypt it
before passing it off to Salt. This allows you to safely store secrets in
source control, in such a way that only your Salt master can decrypt them and
distribute them only to the minions that need them.

The typical use-case would be to use ciphers in your pillar data, and keep a
secret key on your master. You can put the public key in source control so that
developers can add new secrets quickly and easily.

This renderer requires the libsodium library binary and libnacl >= 1.5.1
python package (support for sealed boxes came in 1.5.1 version).


Setup
-----

To set things up, first generate a keypair. On the master, run the following:

.. code-block:: bash

    # salt-call --local nacl.keygen sk_file=/root/.nacl


Using encrypted pillar
----------------------

To encrypt secrets, copy the public key to your local machine and run:

.. code-block:: bash

    $ salt-call --local nacl.enc datatoenc pk_file=/root/.nacl.pub


To apply the renderer on a file-by-file basis add the following line to the
top of any pillar with nacl encrypted data in it:

.. code-block:: yaml

    #!yaml|nacl

Now with your renderer configured, you can include your ciphers in your pillar
data like so:

.. code-block:: yaml

    #!yaml|nacl

    a-secret: "NACL[MRN3cc+fmdxyQbz6WMF+jq1hKdU5X5BBI7OjK+atvHo1ll+w1gZ7XyWtZVfq9gK9rQaMfkDxmidJKwE0Mw==]"
"""


import logging
import re

import salt.syspaths
import salt.utils.stringio

log = logging.getLogger(__name__)
NACL_REGEX = r"^NACL\[(.*)\]$"


def _decrypt_object(obj, **kwargs):
    """
    Recursively try to decrypt any object. If the object is a str, and it
    contains a valid NACLENC pretext, decrypt it, otherwise keep going until a
    string is found.
    """
    if salt.utils.stringio.is_readable(obj):
        return _decrypt_object(obj.getvalue(), **kwargs)
    if isinstance(obj, str):
        if re.search(NACL_REGEX, obj) is not None:
            return __salt__["nacl.dec"](re.search(NACL_REGEX, obj).group(1), **kwargs)
        else:
            return obj
    elif isinstance(obj, dict):
        for key, value in obj.items():
            obj[key] = _decrypt_object(value, **kwargs)
        return obj
    elif isinstance(obj, list):
        for key, value in enumerate(obj):
            obj[key] = _decrypt_object(value, **kwargs)
        return obj
    else:
        return obj


def render(nacl_data, saltenv="base", sls="", argline="", **kwargs):
    """
    Decrypt the data to be rendered using the given nacl key or the one given
    in config
    """
    return _decrypt_object(nacl_data, **kwargs)
